# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import re

import pytest

from pex.pip.version import PipVersion
from pex.typing import TYPE_CHECKING
from testing import PY_VER
from testing.cli import run_pex3

if TYPE_CHECKING:
    from typing import Any


EXPECTED_ERROR_HEADER_LEAD_IN = (
    "Found 7 invalid Requires-Dist metadata values in mlflow 1.27 metadata from "
    "mlflow-1.27.0.dist-info/METADATA at"
)

EXPECTED_WHEEL = "mlflow-1.27.0-py3-none-any.whl"

EXPECTED_ERROR_FOOTER = """\
1. "scikit-learn (>=1.0.*) ; extra == 'pipelines'": .* suffix can only be used with `==` or `!=` operators
    scikit-learn (>=1.0.*) ; extra == 'pipelines'
                  ~~~~~~^
2. "pyarrow (>=7.0.*) ; extra == 'pipelines'": .* suffix can only be used with `==` or `!=` operators
    pyarrow (>=7.0.*) ; extra == 'pipelines'
             ~~~~~~^
3. "shap (>=0.40.*) ; extra == 'pipelines'": .* suffix can only be used with `==` or `!=` operators
    shap (>=0.40.*) ; extra == 'pipelines'
          ~~~~~~~^
4. "pandas-profiling (>=3.1.*) ; extra == 'pipelines'": .* suffix can only be used with `==` or `!=` operators
    pandas-profiling (>=3.1.*) ; extra == 'pipelines'
                      ~~~~~~^
5. "ipython (>=7.0.*) ; extra == 'pipelines'": .* suffix can only be used with `==` or `!=` operators
    ipython (>=7.0.*) ; extra == 'pipelines'
             ~~~~~~^
6. "markdown (>=3.3.*) ; extra == 'pipelines'": .* suffix can only be used with `==` or `!=` operators
    markdown (>=3.3.*) ; extra == 'pipelines'
              ~~~~~~^
7. "Jinja2 (>=3.0.*) ; extra == 'pipelines'": .* suffix can only be used with `==` or `!=` operators
    Jinja2 (>=3.0.*) ; extra == 'pipelines'
            ~~~~~~^
"""


@pytest.mark.skipif(
    PY_VER < (3, 7) or PipVersion.DEFAULT.version >= PipVersion.v24_1.version,
    reason="The mlflow 1.27.0 distribution requires Python>=3.7 and the test requires Pip<24.1.",
)
def test_invalid_metadata_error_messages_under_old_pip(tmpdir):
    # type: (Any) -> None

    run_pex3("lock", "create", "mlflow==1.27.0", "--intransitive").assert_failure(
        expected_error_re=r"^{header_lead_in} .*/{expected_wheel}:\n{footer}$".format(
            header_lead_in=re.escape(EXPECTED_ERROR_HEADER_LEAD_IN),
            expected_wheel=re.escape(EXPECTED_WHEEL),
            footer=re.escape(EXPECTED_ERROR_FOOTER),
        )
    )



# Copyright 2020 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import sys

from pex.tools.main import main as tools

sys.exit(tools())



# Copyright 2020 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, print_function

import os
import re
import tarfile
import warnings
from contextlib import contextmanager
from textwrap import dedent

import pytest

from pex.common import open_zip, safe_open, temporary_dir, touch
from pex.dist_metadata import (
    Distribution,
    InvalidMetadataError,
    MetadataError,
    MetadataType,
    ProjectNameAndVersion,
    Requirement,
    project_name_and_version,
    requires_dists,
    requires_python,
)
from pex.pep_427 import install_wheel_chroot
from pex.pep_503 import ProjectName
from pex.pex_warnings import PEXWarning
from pex.pip.installation import get_pip
from pex.resolve.resolver_configuration import BuildConfiguration
from pex.third_party.packaging.specifiers import SpecifierSet
from pex.typing import TYPE_CHECKING
from testing import PY_VER
from testing.dist_metadata import create_dist_metadata

if TYPE_CHECKING:
    from typing import Any, Iterator, Tuple


@contextmanager
def installed_wheel(wheel_path):
    # type: (str) -> Iterator[Distribution]
    with temporary_dir() as install_dir:
        install_wheel_chroot(wheel_path=wheel_path, destination=install_dir)
        yield Distribution.load(install_dir)


@contextmanager
def example_distribution(name):
    # type: (str) -> Iterator[Tuple[str, Distribution]]
    wheel_path = os.path.join("./tests/example_packages", name)
    with installed_wheel(wheel_path) as distribution:
        yield wheel_path, distribution


@contextmanager
def downloaded_sdist(requirement):
    # type: (str) -> Iterator[str]
    with temporary_dir() as td:
        download_dir = os.path.join(td, "download")
        get_pip().spawn_download_distributions(
            download_dir=download_dir,
            requirements=[requirement],
            transitive=False,
            build_configuration=BuildConfiguration.create(allow_wheels=False),
        ).wait()
        dists = os.listdir(download_dir)
        assert len(dists) == 1, "Expected 1 dist to be downloaded for {}.".format(requirement)
        sdist = os.path.join(download_dir, dists[0])
        assert sdist.endswith((".sdist", ".tar.gz", ".zip"))
        yield sdist


def as_requirement(project_name_and_version):
    # type: (ProjectNameAndVersion) -> str
    return "{}=={}".format(project_name_and_version.project_name, project_name_and_version.version)


PYGOOGLEEARTH_PROJECT_NAME_AND_VERSION = ProjectNameAndVersion("pygoogleearth", "0.0.2")


@pytest.fixture(scope="module")
def pygoogleearth_zip_sdist():
    # type: () -> Iterator[str]
    with downloaded_sdist(as_requirement(PYGOOGLEEARTH_PROJECT_NAME_AND_VERSION)) as sdist:
        assert sdist.endswith(".zip")
        yield sdist


PIP_PROJECT_NAME_AND_VERSION = ProjectNameAndVersion("pip", "9.0.1")


@pytest.fixture(scope="module")
def pip_tgz_sdist():
    # type: () -> Iterator[str]
    with downloaded_sdist(as_requirement(PIP_PROJECT_NAME_AND_VERSION)) as sdist:
        assert sdist.endswith(".tar.gz")
        yield sdist


@pytest.fixture(scope="module")
def pip_wheel(pip_tgz_sdist):
    # type: (str) -> Iterator[str]
    with temporary_dir() as wheel_dir:
        get_pip().spawn_build_wheels([pip_tgz_sdist], wheel_dir=wheel_dir).wait()
        wheels = os.listdir(wheel_dir)
        assert len(wheels) == 1, "Expected 1 wheel to be built for {}.".format(pip_tgz_sdist)
        wheel = os.path.join(wheel_dir, wheels[0])
        assert wheel.endswith(".whl")
        yield wheel


@pytest.fixture(scope="module")
def pip_distribution(pip_wheel):
    # type: (str) -> Iterator[Distribution]
    with installed_wheel(pip_wheel) as distribution:
        yield distribution


def test_project_name_and_version_from_filename(
    pygoogleearth_zip_sdist,  # type: str
    pip_tgz_sdist,  # type: str
    pip_wheel,  # type: str
):
    # type: (...) -> None
    assert PYGOOGLEEARTH_PROJECT_NAME_AND_VERSION == ProjectNameAndVersion.from_filename(
        pygoogleearth_zip_sdist
    )
    assert PIP_PROJECT_NAME_AND_VERSION == ProjectNameAndVersion.from_filename(pip_tgz_sdist)
    assert PIP_PROJECT_NAME_AND_VERSION == ProjectNameAndVersion.from_filename(pip_wheel)


def test_project_name_and_version_from_filename_pep625():
    # type: () -> None
    assert ProjectNameAndVersion(
        "a-distribution-name", "1.2.3"
    ) == ProjectNameAndVersion.from_filename("a-distribution-name-1.2.3.tar.gz")


def test_project_name_and_version_from_filename_invalid():
    # type: () -> None
    with pytest.raises(MetadataError):
        ProjectNameAndVersion.from_filename("unknown_distribution.format")


def test_project_name_and_version_from_metadata(
    pygoogleearth_zip_sdist,  # type: str
    pip_tgz_sdist,  # type: str
    pip_wheel,  # type: str
    pip_distribution,  # type: Distribution
):
    # type: (...) -> None
    assert PYGOOGLEEARTH_PROJECT_NAME_AND_VERSION == project_name_and_version(
        pygoogleearth_zip_sdist, fallback_to_filename=False
    )
    assert PIP_PROJECT_NAME_AND_VERSION == project_name_and_version(
        pip_tgz_sdist, fallback_to_filename=False
    )
    assert PIP_PROJECT_NAME_AND_VERSION == project_name_and_version(
        pip_wheel, fallback_to_filename=False
    )
    assert PIP_PROJECT_NAME_AND_VERSION == project_name_and_version(
        pip_distribution, fallback_to_filename=False
    )


def test_project_name_and_version_fallback(tmpdir):
    # type: (Any) -> None
    def tmp_path(relpath):
        # type: (str) -> str
        return os.path.join(str(tmpdir), relpath)

    expected_metadata_project_name_and_version = ProjectNameAndVersion("foo", "1.2.3")

    pkg_info_src = tmp_path("PKG-INFO")
    with open(pkg_info_src, "w") as fp:
        fp.write("Name: {}\n".format(expected_metadata_project_name_and_version.project_name))
        fp.write("Version: {}\n".format(expected_metadata_project_name_and_version.version))

    sdist_path = tmp_path("bar-baz-4.5.6.tar.gz")
    with tarfile.open(sdist_path, mode="w:gz") as tf:
        # N.B.: Valid PKG-INFO at an invalid location.
        tf.add(pkg_info_src, arcname="PKG-INFO")

    assert project_name_and_version(sdist_path, fallback_to_filename=False) is None
    assert ProjectNameAndVersion("bar-baz", "4.5.6") == project_name_and_version(
        sdist_path, fallback_to_filename=True
    )

    name_and_version = "eggs-7.8.9"
    pkf_info_path = "{}/PKG-INFO".format(name_and_version)

    def write_sdist_tgz(extension):
        # type: (str) -> str
        sdist_path = tmp_path("{}.{}".format(name_and_version, extension))
        with tarfile.open(sdist_path, mode="w:gz") as tf:
            tf.add(pkg_info_src, arcname=pkf_info_path)
        return sdist_path

    assert expected_metadata_project_name_and_version == project_name_and_version(
        write_sdist_tgz("tar.gz"), fallback_to_filename=False
    )
    assert expected_metadata_project_name_and_version == project_name_and_version(
        write_sdist_tgz("tgz"), fallback_to_filename=False
    )

    zip_sdist_path = tmp_path("{}.zip".format(name_and_version))
    with open_zip(zip_sdist_path, mode="w") as zf:
        zf.write(pkg_info_src, arcname=pkf_info_path)

    assert expected_metadata_project_name_and_version == project_name_and_version(
        zip_sdist_path, fallback_to_filename=False
    )


def test_requires_python(
    pip_tgz_sdist,  # type: str
    pip_wheel,  # type: str
    pip_distribution,  # type: Distribution
):
    # type: (...) -> None
    expected_requires_python = SpecifierSet(">=2.6,!=3.0.*,!=3.1.*,!=3.2.*")
    assert expected_requires_python == requires_python(pip_tgz_sdist)
    assert expected_requires_python == requires_python(pip_wheel)
    assert expected_requires_python == requires_python(pip_distribution)


def test_requires_python_none(pygoogleearth_zip_sdist):
    # type: (str) -> None
    assert requires_python(pygoogleearth_zip_sdist) is None
    with example_distribution("aws_cfn_bootstrap-1.4-py2-none-any.whl") as (wheel_path, dist):
        assert requires_python(wheel_path) is None
        assert requires_python(dist) is None


def test_requires_dists():
    # type: () -> None
    with example_distribution("aws_cfn_bootstrap-1.4-py2-none-any.whl") as (wheel_path, dist):
        expected_requirements = [
            Requirement.parse(req)
            for req in ("python-daemon>=1.5.2,<2.0", "pystache>=0.4.0", "setuptools")
        ]
        assert expected_requirements == list(requires_dists(wheel_path))
        assert expected_requirements == list(requires_dists(dist))


def test_requires_dists_none(pygoogleearth_zip_sdist):
    # type: (str) -> None
    assert [] == list(requires_dists(pygoogleearth_zip_sdist))
    with example_distribution("MarkupSafe-1.0-cp27-cp27mu-linux_x86_64.whl") as (wheel_path, dist):
        assert [] == list(requires_dists(wheel_path))
        assert [] == list(requires_dists(dist))

    # This tests a strange case detailed here:
    #   https://github.com/pex-tool/pex/issues/1201#issuecomment-791715585
    with downloaded_sdist("et-xmlfile==1.0.1") as sdist, warnings.catch_warnings(
        record=True
    ) as events:
        assert [] == list(requires_dists(sdist))
        assert len(events) == 1
        warning = events[0]
        assert PEXWarning == warning.category
        assert (
            dedent(
                """\
                Ignoring 1 `Requires` field in {sdist} metadata:
                1.) Requires: python (>=2.6.0)

                You may have issues using the 'et-xmlfile' distribution as a result.
                More information on this workaround can be found here:
                  https://github.com/pex-tool/pex/issues/1201#issuecomment-791715585
                """
            ).format(sdist=sdist)
            == str(warning.message)
        )


def test_wheel_metadata_project_name_fuzzy_issues_1375():
    # type: () -> None
    with example_distribution("PyAthena-1.9.0-py2.py3-none-any.whl") as (wheel_path, dist):
        expected = ProjectNameAndVersion("PyAthena", "1.9.0")
        assert expected == project_name_and_version(wheel_path)
        assert expected == project_name_and_version(dist)

    with example_distribution("PyAthena-1.11.5-py2.py3-none-any.whl") as (wheel_path, dist):
        expected = ProjectNameAndVersion("pyathena", "1.11.5")
        assert expected == project_name_and_version(wheel_path)
        assert expected == project_name_and_version(dist)


@pytest.mark.parametrize(
    "metadata_type",
    [
        pytest.param(metadata_type, id=str(metadata_type))
        for metadata_type in (MetadataType.DIST_INFO, MetadataType.EGG_INFO)
    ],
)
def test_find_dist_info_file(
    tmpdir,  # type: Any
    metadata_type,  # type: MetadataType.Value
):
    # type: (...) -> None
    assert (
        metadata_type.load_metadata(location=str(tmpdir), project_name=ProjectName("foo")) is None
    )

    def metadata_dir_name(project_name_and_version):
        # type: (str) -> str
        return "{project_name_and_version}.{metadata_type}".format(
            project_name_and_version=project_name_and_version,
            metadata_type="dist-info" if metadata_type is MetadataType.DIST_INFO else "egg-info",
        )

    metadata_file_name = "METADATA" if metadata_type is MetadataType.DIST_INFO else "PKG-INFO"

    touch(os.path.join(str(tmpdir), metadata_dir_name("foo-1.0"), "baz"))
    assert (
        metadata_type.load_metadata(location=str(tmpdir), project_name=ProjectName("foo")) is None
    )

    touch(os.path.join(str(tmpdir), metadata_dir_name("foo-1.0"), metadata_file_name))
    assert (
        metadata_type.load_metadata(location=str(tmpdir), project_name=ProjectName("foo")) is None
    )

    def write_pkg_info_file(
        location,  # type: str
        name,  # type: str
        version,  # type: str
    ):
        # type: (...) -> None
        with safe_open(
            os.path.join(
                location,
                metadata_dir_name("{name}-{version}".format(name=name, version=version)),
                metadata_file_name,
            ),
            "w",
        ) as fp:
            print("Metadata-Version: 1.0", file=fp)
            print("Name: {name}".format(name=name), file=fp)
            print("Version: {version}".format(version=version), file=fp)

    foo_location = os.path.join(str(tmpdir), "foo_location")
    touch(os.path.join(foo_location, metadata_dir_name("foo-100"), "bar"))
    expected_metadata_relpath = os.path.join(metadata_dir_name("Foo-1.0"), "bar")
    touch(os.path.join(foo_location, expected_metadata_relpath))
    write_pkg_info_file(foo_location, name="Foo", version="1.0")

    metadata_files = metadata_type.load_metadata(
        location=foo_location, project_name=ProjectName("foo")
    )
    assert metadata_files is not None
    assert expected_metadata_relpath == metadata_files.metadata_file_rel_path("bar")

    stress_location = os.path.join(str(tmpdir), "stress_location")
    touch(os.path.join(stress_location, "direct_url.json"))
    touch(os.path.join(stress_location, metadata_dir_name("foo-1.0rc0"), "direct_url.json"))
    expected_metadata_relpath = os.path.join(
        metadata_dir_name("stress__-.-__Test-1.0rc0"), "direct_url.json"
    )
    touch(os.path.join(stress_location, expected_metadata_relpath))
    write_pkg_info_file(
        stress_location,
        name="stress__-.-__Test",
        version="1.0rc0",
    )

    metadata_files = metadata_type.load_metadata(
        location=stress_location, project_name=ProjectName("Stress-.__Test")
    )
    assert metadata_files is not None
    assert expected_metadata_relpath == metadata_files.metadata_file_rel_path("direct_url.json")


def test_requirement_contains_requirement():
    req = Requirement.parse

    assert req("foo==1") in req("foo")
    assert req("foo") not in req("foo==1")

    assert req("bar") not in req("foo")
    assert req("bar==1") not in req("foo==1")

    assert req("foo>1,<3") not in req("foo==2")
    assert req("foo==2") in req("foo>1,<3")
    assert req("foo>=1,<3") not in req("foo>1,<3")
    assert req("foo>1,<=2") in req("foo>1,<3")


def test_invalid_metadata_requires_python_error():
    with pytest.raises(
        InvalidMetadataError,
        match=re.escape(
            "Invalid Requires-Python metadata found in foo 0.1 metadata from "
            "foo-0.1.0.dist-info/METADATA at /right/here 'not a valid specifier': "
            "Invalid specifier: 'not a valid specifier'"
        ),
    ):
        create_dist_metadata(
            project_name="foo",
            version="0.1.0",
            requires_python="not a valid specifier",
            location="/right/here",
        )


@pytest.mark.skipif(
    PY_VER < (3, 7),
    reason=(
        "The vendored packaging for Python < 3.7 does not complain about use of `*` for operators "
        "other than `==` and `!=`."
    ),
)
def test_invalid_metadata_requires_dists_error_issue_2441(tmpdir):
    with pytest.raises(
        InvalidMetadataError,
        match=re.escape(
            "Found 2 invalid Requires-Dist metadata values in foo 0.1 metadata from "
            "foo-0.1.0.dist-info/METADATA at /right/here:\n"
            "1. \"scikit-learn (>=1.0.*) ; extra == 'pipelines'\": "
            ".* suffix can only be used with `==` or `!=` operators\n"
            "    scikit-learn (>=1.0.*) ; extra == 'pipelines'\n"
            "                  ~~~~~~^\n"
            "2. \"pyarrow (>=7.0.*) ; extra == 'pipelines'\": "
            ".* suffix can only be used with `==` or `!=` operators\n"
            "    pyarrow (>=7.0.*) ; extra == 'pipelines'\n"
            "             ~~~~~~^"
        ),
    ):
        create_dist_metadata(
            project_name="foo",
            version="0.1.0",
            requires_dists=(
                "cowsay<6",
                "scikit-learn (>=1.0.*) ; extra == 'pipelines'",
                "ansicolors==1.1.8",
                "pyarrow (>=7.0.*) ; extra == 'pipelines'",
            ),
            location="/right/here",
        )

    location = str(tmpdir)
    with safe_open(os.path.join(location, "foo-0.1.0.egg-info", "requires.txt"), "w") as fp:
        fp.write(
            dedent(
                """\
                cowsay<6
                ansicolors==1.1.8

                [pipelines]
                scikit-learn>=1.0.*
                pyarrow>=7.0.*
                """
            )
        )
    with pytest.raises(
        InvalidMetadataError,
        match=re.escape(
            "Found 2 invalid Requires-Dist metadata values in foo 0.1 metadata from "
            "foo-0.1.0.egg-info/PKG-INFO at {location}:\n"
            "1. foo-0.1.0.egg-info/requires.txt:5 'scikit-learn>=1.0.*; extra == \"pipelines\"': "
            ".* suffix can only be used with `==` or `!=` operators\n"
            '    scikit-learn>=1.0.*; extra == "pipelines"\n'
            "                ~~~~~~^\n"
            "2. foo-0.1.0.egg-info/requires.txt:6 'pyarrow>=7.0.*; extra == \"pipelines\"': "
            ".* suffix can only be used with `==` or `!=` operators\n"
            '    pyarrow>=7.0.*; extra == "pipelines"'
            "\n           ~~~~~~^".format(location=location)
        ),
    ):
        create_dist_metadata(
            project_name="foo",
            version="0.1.0",
            location=location,
            metadata_type=MetadataType.EGG_INFO,
        )



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import filecmp
import os.path
import re
import shutil
import subprocess
from textwrap import dedent

from colors import color

from pex.common import safe_open, touch
from pex.targets import LocalInterpreter
from pex.typing import TYPE_CHECKING
from testing import run_pex_command
from testing.cli import run_pex3

if TYPE_CHECKING:
    from typing import Any


def test_invalid_project(
    tmpdir,  # type: Any
    pex_project_dir,  # type: str
):
    # type: (...) -> None

    non_project_dir = os.path.join(str(tmpdir), "non-project-dir")
    os.mkdir(non_project_dir)
    run_pex_command(
        args=["--project", pex_project_dir, "--project", non_project_dir]
    ).assert_failure(
        expected_error_re=r".*{message}$".format(
            message=re.escape(
                "The following --project paths do not appear to point to directories containing "
                "Python projects:\n"
                "1. {non_project_dir}".format(non_project_dir=non_project_dir)
            )
        ),
        re_flags=re.DOTALL,
    )

    non_project_file = os.path.join(str(tmpdir), "non-project-file")
    touch(non_project_file)
    run_pex3(
        "lock",
        "create",
        "--project",
        non_project_dir,
        "--project",
        pex_project_dir,
        "--project",
        non_project_file,
    ).assert_failure(
        expected_error_re=r".*{message}$".format(
            message=re.escape(
                "The following --project paths do not appear to point to directories containing "
                "Python projects:\n"
                "1. {non_project_dir}\n"
                "2. {non_project_file}".format(
                    non_project_dir=non_project_dir, non_project_file=non_project_file
                )
            )
        ),
        re_flags=re.DOTALL,
    )


def test_locked_project(tmpdir):
    # type: (Any) -> None

    project_dir = os.path.join(str(tmpdir), "project")

    def write_speak(fg_color):
        # type: (str) -> None
        with safe_open(os.path.join(project_dir, "speak.py"), "w") as fp:
            fp.write(
                dedent(
                    """\
                    import cowsay
                    from colors import color


                    def tux():
                        cowsay.tux(color("Moo?", fg={fg_color!r}))
                    """
                ).format(fg_color=fg_color)
            )

    def write_setup(cowsay_requirement):
        # type: (str) -> None
        with safe_open(os.path.join(project_dir, "setup.py"), "w") as fp:
            fp.write(
                dedent(
                    """\
                    from setuptools import setup


                    setup(
                        name="speak",
                        version="0.1",
                        install_requires=["ansicolors", {cowsay_requirement!r}],
                        entry_points={{
                            "console_scripts": [
                                "speak = speak:tux",
                            ],
                        }},
                        py_modules=["speak"],
                    )
                    """
                ).format(cowsay_requirement=cowsay_requirement)
            )

    def assert_pex(
        pex,  # type: str
        expected_color,  # type: str
    ):
        # type: (...) -> None
        assert "| {message} |".format(
            message=color("Moo?", fg=expected_color)
        ) in subprocess.check_output(args=[pex]).decode("utf-8")

    pex_root = os.path.join(str(tmpdir), "pex-root")

    write_speak(fg_color="brown")
    write_setup(cowsay_requirement="cowsay<5")

    project_lock = os.path.join(str(tmpdir), "project-lock.json")
    run_pex3(
        "lock", "create", "--pex-root", pex_root, project_dir, "--indent", "2", "-o", project_lock
    ).assert_success()
    pex1 = os.path.join(str(tmpdir), "pex1")
    run_pex_command(
        args=[
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
            "--lock",
            project_lock,
            "-c",
            "speak",
            "-o",
            pex1,
        ]
    ).assert_success()
    assert_pex(pex1, expected_color="brown")

    third_party_lock = os.path.join(str(tmpdir), "third-party-lock.json")
    run_pex3(
        "lock",
        "create",
        "--pex-root",
        pex_root,
        "--project",
        project_dir,
        "--indent",
        "2",
        "-o",
        third_party_lock,
    ).assert_success()
    pex2 = os.path.join(str(tmpdir), "pex2")
    run_pex_command(
        args=[
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
            "--project",
            project_dir,
            "--lock",
            third_party_lock,
            "-c",
            "speak",
            "-o",
            pex2,
        ]
    ).assert_success()
    assert_pex(pex2, expected_color="brown")

    # Using `pex --project local/project` should produce identical results to `pex local/project`;
    # the utility of `pex --project` comes only when combined with a lock or PEX repository.
    assert filecmp.cmp(pex1, pex2, shallow=False)

    # Modifying the project should invalidate the full project lock, but work with a `--project`
    # lock.
    write_speak(fg_color="blue")
    shutil.rmtree(pex_root)

    run_pex_command(
        args=[
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
            "--lock",
            project_lock,
            "-c",
            "speak",
            "-o",
            pex1,
        ]
    ).assert_failure(
        expected_error_re=(
            r".*{lead_in_message} {sha256_re} when downloading speak but hashed to "
            r"{sha256_re}\.\n$".format(
                lead_in_message=re.escape(
                    "There was 1 error downloading required artifacts:\n"
                    "1. speak 0.1 from file://{project_dir}\n"
                    "    Expected sha256 hash of".format(project_dir=project_dir)
                ),
                sha256_re=r"[a-f0-9]{64}",
            )
        ),
        re_flags=re.DOTALL,
    )

    pex3 = os.path.join(str(tmpdir), "pex3")
    run_pex_command(
        args=[
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
            "--project",
            project_dir,
            "--lock",
            third_party_lock,
            "-c",
            "speak",
            "-o",
            pex3,
        ]
    ).assert_success()
    assert_pex(pex3, expected_color="blue")

    # If the project is updated in a way incompatible with the lock, building a
    # `pex --project ... --lock ...` should fail.
    write_setup(cowsay_requirement="cowsay==5.0")
    target = LocalInterpreter.create()
    run_pex_command(
        args=[
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
            "--project",
            project_dir,
            "--lock",
            third_party_lock,
            "-c",
            "speak",
            "-o",
            pex3,
        ]
    ).assert_failure(
        expected_error_re=r".*{message}$".format(
            message=re.escape(
                "Failed to resolve compatible artifacts from lock {lock} for 1 target:\n"
                "1. {target}:\n"
                "    Failed to resolve all requirements for {target_description} from {lock}:\n"
                "\n"
                "Configured with:\n"
                "    build: True\n"
                "    use_wheel: True\n"
                "\n"
                "Dependency on cowsay not satisfied, 1 incompatible candidate found:\n"
                "1.) cowsay 4 does not satisfy the following requirements:\n"
                "    ==5.0 (via: cowsay==5.0)\n".format(
                    lock=third_party_lock,
                    target=target,
                    target_description=target.render_description(),
                )
            )
        ),
        re_flags=re.DOTALL,
    )

    # But syncing the `--project` lock to the modified project should re-right the ship.
    run_pex3(
        "lock",
        "sync",
        "--pex-root",
        pex_root,
        "--project",
        project_dir,
        "--indent",
        "2",
        "--lock",
        third_party_lock,
    ).assert_success()
    run_pex_command(
        args=[
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
            "--project",
            project_dir,
            "--lock",
            third_party_lock,
            "-c",
            "speak",
            "-o",
            pex3,
        ]
    ).assert_success()
    assert_pex(pex3, expected_color="blue")



# Copyright 2023 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path

import pytest

from pex.interpreter_constraints import InterpreterConstraint
from pex.pep_440 import Version
from pex.pip.version import PipVersion, PipVersionValue
from pex.typing import TYPE_CHECKING
from testing import run_pex_command
from testing.cli import run_pex3

if TYPE_CHECKING:
    from typing import Any


# N.B.: awscli==1.28.1 Only resolves with the pip-2020-resolver for Pip versions earlier than 22.
# In the end this means only Pip >=20.3.2,<22 can run this test successfully. The underlying issue
# comes from PyYAML and is documented here: https://github.com/yaml/pyyaml/issues/601
@pytest.mark.parametrize(
    "pip_version",
    [
        pytest.param(pip_version, id=str(pip_version))
        for pip_version in PipVersion.values()
        # MyPy fails to typecheck <= under Python 2.7 only, even though Version has @total_ordering
        # applied.
        if Version("20.3.2") <= pip_version.version < Version("22")  # type: ignore[operator]
    ],
)
def test_backtracking(
    tmpdir,  # type: Any
    pip_version,  # type: PipVersionValue
):
    # type: (...) -> None

    lock = os.path.join(str(tmpdir), "lock.json")
    run_pex3(
        "lock",
        "create",
        "-v",
        "-o",
        lock,
        "--indent",
        "2",
        "--pip-version",
        str(pip_version),
        "--resolver-version",
        "pip-2020-resolver",
        "--interpreter-constraint",
        "CPython==3.11.*",
        "--style",
        "universal",
        "--target-system",
        "linux",
        "--target-system",
        "mac",
        "--manylinux",
        "manylinux2014",
        "awscli==1.28.1",
    ).assert_success()

    try:
        python311 = next(InterpreterConstraint.parse("CPython==3.11.*").iter_matching())
    except StopIteration:
        pytest.skip("Skipping lock verification since no CPython 3.11 interpreter is available.")

    result = run_pex_command(
        args=[
            "--lock",
            lock,
            "-c",
            "aws",
            "--",
            "--version",
        ],
        python=python311.binary,
    )
    result.assert_success()
    assert result.output.startswith("aws-cli/1.28.1 "), result.output



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import subprocess

from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.resolve.locked_resolve import FileArtifact
from pex.resolve.lockfile import json_codec
from pex.resolve.resolved_requirement import Pin
from pex.typing import TYPE_CHECKING
from testing import run_pex_command
from testing.cli import run_pex3

if TYPE_CHECKING:
    from typing import Any


def test_tar_bz2(tmpdir):
    # type: (Any) -> None

    lock = os.path.join(str(tmpdir), "lock.json")
    pex_root = os.path.join(str(tmpdir), "pex_root")
    run_pex3(
        "lock",
        "create",
        "--pex-root",
        pex_root,
        "python-constraint==1.4.0",
        "-o",
        lock,
        "--indent",
        "2",
    ).assert_success()

    lock_file = json_codec.load(lock)
    assert len(lock_file.locked_resolves) == 1

    locked_resolve = lock_file.locked_resolves[0]
    assert len(locked_resolve.locked_requirements) == 1

    locked_requirement = locked_resolve.locked_requirements[0]
    assert Pin(ProjectName("python-constraint"), Version("1.4.0")) == locked_requirement.pin
    assert isinstance(locked_requirement.artifact, FileArtifact)
    assert locked_requirement.artifact.is_source
    assert locked_requirement.artifact.filename.endswith(".tar.bz2")
    assert not locked_requirement.additional_artifacts

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(
        args=["--pex-root", pex_root, "--runtime-pex-root", pex_root, "--lock", lock, "-o", pex]
    ).assert_success()

    assert (
        b"1.4.0"
        == subprocess.check_output(
            args=[pex, "-c", "from constraint.version import __version__; print(__version__)"]
        ).strip()
    )



# SPDX-License-Identifier: MIT

if "attrs" in __import__("os").environ.get("__PEX_UNVENDORED__", ""):
  from attr.converters import *    # vendor:skip
else:
  from pex.third_party.attr.converters import *  
# noqa



# Copyright 2023 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import subprocess
from textwrap import dedent

import pytest

from pex.common import safe_open
from pex.compatibility import commonpath
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.pex import PEX
from pex.typing import TYPE_CHECKING
from pex.venv.virtualenv import Virtualenv
from testing import PY_VER, make_env, run_pex_command

if TYPE_CHECKING:
    from typing import Any


@pytest.mark.skipif(
    PY_VER < (3, 7) or PY_VER >= (3, 12),
    reason=(
        "The test requires use of attrs 23.1.0 which requires Python >= 3.7 and Lambdex further "
        "requires a released version of Pex that supports Python 3.12."
    ),
)
def test_lambdex_with_incompatible_attrs(tmpdir):
    # type: (Any) -> None

    src = os.path.join(str(tmpdir), "src")
    with safe_open(os.path.join(src, "example.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                import sys

                from attr import AttrsInstance

                def run():
                    print(sys.modules[AttrsInstance.__module__].__file__)
                """
            )
        )

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(args=["-D", src, "attrs==23.1.0", "-o", pex]).assert_success()

    pex_distributions_by_project_name = {
        dist.metadata.project_name: dist for dist in PEX(pex).resolve()
    }
    user_attrs = pex_distributions_by_project_name[ProjectName("attrs")]
    assert Version("23.1.0") == user_attrs.metadata.version

    lambda_zip = os.path.join(str(tmpdir), "lambda.zip")
    run_pex_command(
        args=[
            # The Lambdex 0.2.0 final release changed from a top-level pex.third_party import to a
            # lazy one which foils scrubbing under Python<3.8. Instead of releasing a Lambdex 0.2.1
            # with a fix for this eagerly, we just pin this test dep low to avoid the issue and see
            # if any bug report ever comes in, which seems unlikely since Python 3.7 is EOL and
            # Lambdex is as well.
            "lambdex<0.2.0",
            "-c",
            "lambdex",
            "--",
            "build",
            "-e",
            "example:run",
            "-o",
            lambda_zip,
            pex,
        ]
    ).assert_success()

    venv_dir = os.path.join(str(tmpdir), "venv_dir")
    venv = Virtualenv.create(venv_dir=venv_dir)
    output = (
        subprocess.check_output(
            args=[venv.interpreter.binary, "-c", "from lambdex_handler import handler; handler()"],
            env=make_env(PYTHONPATH=lambda_zip),
        )
        .decode("utf-8")
        .strip()
    )
    assert user_attrs.location == commonpath((user_attrs.location, output)), output



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import re
import shutil
from textwrap import dedent
from typing import Dict, Set, Tuple

import pytest

from pex.atomic_directory import atomic_directory
from pex.common import safe_mkdir
from pex.dist_metadata import Requirement
from pex.pep_427 import InstallableType
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.pip.version import PipVersion, PipVersionValue
from pex.resolve.configured_resolver import ConfiguredResolver
from pex.resolve.lockfile import json_codec
from pex.resolve.lockfile.model import Lockfile
from pex.resolve.path_mappings import PathMapping, PathMappings
from pex.resolve.resolved_requirement import Pin
from pex.resolve.resolvers import Resolver
from pex.typing import TYPE_CHECKING
from testing import IntegResults, built_wheel, re_exact
from testing.cli import run_pex3

if TYPE_CHECKING:
    from typing import Any, List, Optional, Protocol

    import attr  # vendor:skip

    class RunLockUpdate(Protocol):
        def __call__(
            self,
            *args,  # type: str
            **popen_kwargs  # type: Any
        ):
            # type: (...) -> IntegResults
            pass

else:
    from pex.third_party import attr


@attr.s(frozen=True)
class FindLinksRepo(object):
    @classmethod
    def create(cls, path):
        # type: (str) -> FindLinksRepo
        safe_mkdir(path, clean=True)
        return cls(path)

    path = attr.ib()  # type: str

    def host(self, distribution):
        # type: (str) -> None
        shutil.copy(distribution, os.path.join(self.path, os.path.basename(distribution)))

    def make_wheel(
        self,
        project_name,  # type: str
        version,  # type: str
        requirements=None,  # type: Optional[List[str]]
    ):
        # type: (...) -> None
        with built_wheel(
            name=project_name, version=version, install_reqs=requirements, universal=True
        ) as wheel:
            self.host(wheel)


def populate_find_links_repo_initial(
    find_links_repo,  # type: FindLinksRepo
    resolver,  # type: Resolver
):
    # type: (...) -> None

    # N.B.: Since we are setting up a find links repo for offline lock resolves, we grab one
    # distribution online to allow the current Pip version to bootstrap itself if needed.
    result = resolver.resolve_requirements(
        ["ansicolors==1.1.8"], result_type=InstallableType.WHEEL_FILE
    )
    for resolved_distribution in result.distributions:
        find_links_repo.host(resolved_distribution.distribution.location)

    find_links_repo.make_wheel("proj-c", "1")
    find_links_repo.make_wheel("proj-b", "1", requirements=["proj-a", "proj-c"])
    find_links_repo.make_wheel("proj-a", "1", requirements=["proj-c", "ansicolors"])
    find_links_repo.make_wheel("proj-a", "1.1", requirements=["proj-c", "ansicolors"])


def populate_find_links_repo_additional(find_links_repo):
    # type: (FindLinksRepo) -> None
    find_links_repo.make_wheel("proj-a", "2", requirements=["ansicolors"])
    find_links_repo.make_wheel("proj-b", "2", requirements=["proj-c"])


_FIND_LINKS_REPO_PATH_MAPPING_NAME = "FL"


@pytest.fixture
def path_mappings(tmpdir):
    # type: (Any) -> PathMappings
    return PathMappings(
        (PathMapping(os.path.join(str(tmpdir), "arbitrary"), _FIND_LINKS_REPO_PATH_MAPPING_NAME),)
    )


@pytest.fixture(scope="session")
def lock_session_fixtures(shared_integration_test_tmpdir):
    # type: (str) -> Tuple[str, PipVersionValue, RunLockUpdate]

    pip_version = PipVersion.DEFAULT

    def repo_args(find_links_dir):
        # type: (str) -> List[str]
        return [
            "--no-pypi",
            "--find-links",
            find_links_dir,
            "--path-mapping",
            "{name}|{find_links}|Test find links repo".format(
                name=_FIND_LINKS_REPO_PATH_MAPPING_NAME, find_links=find_links_dir
            ),
        ]

    test_lock_update_chroot = os.path.join(
        shared_integration_test_tmpdir, "test_lock_update_chroot"
    )
    with atomic_directory(test_lock_update_chroot) as chroot:
        if not chroot.is_finalized():
            resolver = ConfiguredResolver.version(pip_version)

            find_links = os.path.join(chroot.work_dir, "find_links")
            find_links_repo = FindLinksRepo.create(find_links)

            # Control an initial lock.
            populate_find_links_repo_initial(find_links_repo, resolver)
            run_pex3(
                *(
                    [
                        "lock",
                        "create",
                        "--pip-version",
                        str(pip_version),
                        "--resolver-version",
                        "pip-2020-resolver",
                    ]
                    + repo_args(find_links)
                    + [
                        "--style",
                        "universal",
                        "proj-a",
                        "proj-b==1.*",
                        "--indent",
                        "2",
                        "-o",
                        os.path.join(chroot.work_dir, "lock.json"),
                    ]
                )
            ).assert_success()

            # Now advance the state of the work with new versions of projects released.
            populate_find_links_repo_additional(find_links_repo)

    lock = os.path.join(test_lock_update_chroot, "lock.json")

    def run_lock_update(
        *args,  # type: str
        **popenkwargs  # type: Any
    ):
        # type: (...) -> IntegResults
        all_args = ["lock", "update"]
        all_args.extend(repo_args(os.path.join(test_lock_update_chroot, "find_links")))
        all_args.extend(args)
        return run_pex3(*all_args, **popenkwargs)

    return lock, pip_version, run_lock_update


@pytest.fixture
def lock_fixtures(
    tmpdir,  # type: Any
    lock_session_fixtures,  # type: Tuple[str, PipVersionValue, RunLockUpdate]
):
    # type: (...) -> Tuple[str, PipVersionValue, RunLockUpdate]
    lock, pip_version, run_lock_update = lock_session_fixtures

    lock_copy = os.path.join(str(tmpdir), os.path.basename(lock))
    shutil.copy(lock, lock_copy)

    def run_lock_update_on_copy(
        *args,  # type: str
        **popenkwargs  # type: Any
    ):
        # type: (...) -> IntegResults
        all_args = list(args)
        all_args.append(lock_copy)
        return run_lock_update(*all_args, **popenkwargs)

    return lock_copy, pip_version, run_lock_update_on_copy


@pytest.fixture
def lock(lock_fixtures):
    # type: (Tuple[str, PipVersionValue, RunLockUpdate]) -> str
    lock, _, _ = lock_fixtures
    return lock


@pytest.fixture
def pip_version(lock_fixtures):
    # type: (Tuple[str, PipVersionValue, RunLockUpdate]) -> PipVersionValue
    _, pip_version, _ = lock_fixtures
    return pip_version


@pytest.fixture
def run_lock_update(lock_fixtures):
    # type: (Tuple[str, PipVersionValue, RunLockUpdate]) -> RunLockUpdate
    _, _, run_lock_update = lock_fixtures
    return run_lock_update


def test_lock_update_invalid(run_lock_update):
    # type: (RunLockUpdate) -> None
    run_lock_update("-p", "<invalid>").assert_failure(
        expected_error_re=re.escape("Failed to parse project requirement to update '<invalid>': ")
    )


def test_lock_update_nominal(run_lock_update):
    # type: (RunLockUpdate) -> None
    run_lock_update("-p", "proj-a").assert_success(
        expected_error_re=re_exact(
            dedent(
                """\
                Updates for lock generated by universal:
                  Updated proj-a from 1.1 to 2
                """
            )
        )
    )


def test_lock_update_downgrade(run_lock_update):
    # type: (RunLockUpdate) -> None
    run_lock_update("-p", "proj-a<1.1").assert_success(
        expected_error_re=re_exact(
            dedent(
                """\
                Updates for lock generated by universal:
                  Updated proj-a from 1.1 to 1
                Updates to lock input constraints:
                  Added 'proj-a<1.1'
                """
            )
        )
    )


def test_lock_update_conflict(
    run_lock_update,  # type: RunLockUpdate
    pip_version,  # type: PipVersionValue
):
    # type: (...) -> None
    run_lock_update("-p", "proj-b>=2").assert_failure(
        expected_error_re=r".*{header_info}.*{error_message}.*".format(
            header_info=re.escape(
                dedent(
                    """\
                    Given the lock requirements:
                    proj-a
                    proj-b==1.*

                    The following lock update constraints could not all be satisfied:
                    ansicolors==1.1.8
                    proj-a==1.1
                    proj-b>=2
                    proj-c==1
                    """
                )
            ),
            # N.B.: The nice conflict messages are only produced by all the non-vendored Pip versions Pex supports.
            error_message=(
                ""
                if pip_version is PipVersion.VENDORED
                else re.escape(
                    dedent(
                        """\
                        The conflict is caused by:
                             The user requested proj-b==1.*
                             The user requested (constraint) proj-b>=2
                        """
                    )
                )
            ),
        ),
        re_flags=re.DOTALL,
    )


def test_lock_update_replace(run_lock_update):
    # type: (RunLockUpdate) -> None

    # N.B.: This is a conflict specified as a normal project update (see `test_lock_update_conflict` above); the `=`
    # replace directive is what makes this work.
    run_lock_update("-R", "proj-b>=2").assert_success(
        expected_error_re=re_exact(
            dedent(
                """\
                Updates for lock generated by universal:
                  Updated proj-b from 1 to 2
                Updates to lock input requirements:
                  Updated 'proj-b==1.*' to 'proj-b>=2'
                """
            )
        )
    )


def test_lock_update_replace_invalid(run_lock_update):
    # type: (RunLockUpdate) -> None
    run_lock_update("-R", "<invalid>").assert_failure(
        expected_error_re=re.escape("Failed to parse replacement project requirement '<invalid>': ")
    )


def test_lock_update_delete_invalid(run_lock_update):
    # type: (RunLockUpdate) -> None
    run_lock_update("-d", "valid>1.0").assert_failure(
        expected_error_re=re_exact(
            "Failed to parse project name to delete: The given project name 'valid>1.0' is not a valid. "
            "It must conform to the regex '^([A-Z0-9]|[A-Z0-9][A-Z0-9._-]*[A-Z0-9])$' as specified in "
            "https://peps.python.org/pep-0508/#names"
        )
    )


def index_requirements(lockfile):
    # type: (Lockfile) -> Dict[ProjectName, Requirement]
    return {req.project_name: req for req in lockfile.requirements}


def locked_requirement_pins(lockfile):
    # type: (Lockfile) -> Set[Pin]
    assert len(lockfile.locked_resolves) == 1
    locked_resolve = lockfile.locked_resolves[0]
    return {locked_requirement.pin for locked_requirement in locked_resolve.locked_requirements}


def pin(
    project_name,  # type: str
    version,  # type: str
):
    # type: (...) -> Pin
    return Pin(ProjectName(project_name), Version(version))


def test_lock_update_delete_nominal(
    run_lock_update,  # type: RunLockUpdate
    lock,  # type: str
    path_mappings,  # type: PathMappings
):
    # type: (...) -> None

    lockfile = json_codec.load(lock, path_mappings)
    requirements = index_requirements(lockfile)
    proj_a_req = requirements.pop(ProjectName("proj-a"))
    requirements.pop(ProjectName("proj-b"))
    assert not requirements

    run_lock_update("-d", "proj-b").assert_success(
        expected_error_re=re_exact(
            dedent(
                """\
                Updates for lock generated by universal:
                  Deleted proj-b 1
                Updates to lock input requirements:
                  Deleted 'proj-b==1.*'
                """
            )
        )
    )

    lockfile = json_codec.load(lock, path_mappings)
    requirements = index_requirements(lockfile)
    assert proj_a_req == requirements.pop(ProjectName("proj-a"))
    assert not requirements

    locked_requirements = locked_requirement_pins(lockfile)
    assert {
        pin("proj-a", "1.1"),
        pin("proj-c", "1"),
        pin("ansicolors", "1.1.8"),
    } == locked_requirements


def test_lock_update_delete_req_only(
    run_lock_update,  # type: RunLockUpdate
    lock,  # type: str
    path_mappings,  # type: PathMappings
):
    # type: (...) -> None

    lockfile = json_codec.load(lock, path_mappings)
    proj_a_req = index_requirements(lockfile)[ProjectName("proj-a")]
    original_locked_resolves = lockfile.locked_resolves

    run_lock_update("-d", "proj-a").assert_success(
        expected_error_re=re_exact(
            dedent(
                """\
                No updates for lock generated by universal.
                Updates to lock input requirements:
                  Deleted 'proj-a'
                """
            )
        )
    )
    lockfile = json_codec.load(lock, path_mappings)
    assert proj_a_req not in lockfile.requirements
    assert original_locked_resolves == lockfile.locked_resolves


def test_lock_update_delete_noop(
    run_lock_update,  # type: RunLockUpdate
    lock,  # type: str
    path_mappings,  # type: PathMappings
):
    # type: (...) -> None
    original_lockfile = json_codec.load(lock, path_mappings)
    run_lock_update("-d", "proj-c").assert_success(
        expected_error_re=re_exact("No updates for lock generated by universal.")
    )
    assert original_lockfile == json_codec.load(lock, path_mappings)


def test_lock_update_mixed(
    run_lock_update,  # type: RunLockUpdate
    lock,  # type: str
    path_mappings,  # type: PathMappings
):
    # type: (...) -> None

    run_lock_update("-p", "proj-a", "-d", "proj-b").assert_success(
        expected_error_re=re_exact(
            dedent(
                """\
                Updates for lock generated by universal:
                  Deleted proj-b 1
                  Deleted proj-c 1
                  Updated proj-a from 1.1 to 2
                Updates to lock input requirements:
                  Deleted 'proj-b==1.*'
                """
            )
        )
    )
    lockfile = json_codec.load(lock, path_mappings)
    locked_requirements = locked_requirement_pins(lockfile)
    assert {pin("proj-a", "2"), pin("ansicolors", "1.1.8")} == locked_requirements



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import print_function

import os
import sys


def write_bindings(
    env_file,  # type: str
    installed_pex_dir,  # type: str
):
    # type: (...) -> None
    with open(env_file, "a") as fp:
        print("PYTHON=" + sys.executable, file=fp)
        print("PEX=" + os.path.realpath(os.path.join(installed_pex_dir, "__main__.py")), file=fp)


if __name__ == "__main__":
    write_bindings(
        env_file=os.environ["SCIE_BINDING_ENV"],
        installed_pex_dir=(
            # The zipapp case:
            os.environ["_PEX_SCIE_INSTALLED_PEX_DIR"]
            # The --venv case:
            or os.environ.get("VIRTUAL_ENV", os.path.dirname(os.path.dirname(sys.executable)))
        ),
    )
    sys.exit(0)



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from pex.cli.command import BuildTimeCommand
from pex.docs.command import HtmlDocsConfig, register_open_options, serve_html_docs
from pex.docs.server import SERVER_NAME, Pidfile
from pex.docs.server import shutdown as shutdown_docs_server
from pex.result import Ok, Result, try_


class Docs(BuildTimeCommand):
    """Interact with the Pex documentation.

    With no arguments, ensures a local documentation server is running and then opens a browser
    to view the local docs.
    """

    @classmethod
    def add_extra_arguments(cls, parser):
        register_open_options(parser)
        parser.add_argument(
            "--no-open",
            dest="open",
            default=True,
            action="store_false",
            help="Don't open the docs; just ensure the docs server is running and print its info.",
        )
        kill_or_info = parser.add_mutually_exclusive_group()
        kill_or_info.add_argument(
            "-k",
            "--kill-server",
            dest="kill_server",
            default=False,
            action="store_true",
            help="Shut down the {server} if it is running.".format(server=SERVER_NAME),
        )
        kill_or_info.add_argument(
            "--server-info",
            dest="server_info",
            default=False,
            action="store_true",
            help="Print information about the status of the {server}.".format(server=SERVER_NAME),
        )

    def run(self):
        # type: () -> Result

        if self.options.server_info:
            pidfile = Pidfile.load()
            if pidfile and pidfile.alive():
                return Ok(
                    "{server} serving {info}".format(server=SERVER_NAME, info=pidfile.server_info)
                )
            return Ok("No {server} is running.".format(server=SERVER_NAME))

        if self.options.kill_server:
            server_info = shutdown_docs_server()
            if server_info:
                return Ok("Shut down {server} {info}".format(server=SERVER_NAME, info=server_info))
            return Ok("No {server} was running.".format(server=SERVER_NAME))

        launch_result = try_(
            serve_html_docs(
                open_browser=self.options.open, config=HtmlDocsConfig.from_options(self.options)
            )
        )
        if self.options.open:
            return Ok()

        return Ok(
            (
                "{server} already running {info}"
                if launch_result.already_running
                else "Launched {server} {info}"
            ).format(server=SERVER_NAME, info=launch_result.server_info)
        )



# Copyright 2021 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import subprocess

import pytest

from pex.common import temporary_dir
from testing import IS_ARM_64, IS_PYPY, run_pex_command


@pytest.mark.skipif(
    IS_ARM_64 or IS_PYPY,
    reason=(
        "No wheels are published for arm and the sdist fails to compile against modern OpenSSL. "
        "Also, on PyPy we get this error: Failed to execute PEX file. Needed "
        "manylinux2014_x86_64-pp-272-pypy_41 compatible dependencies for 1: "
        "cryptography==2.5 But this pex only contains "
        "cryptography-2.5-pp27-pypy_41-linux_x86_64.whl. "
        "Temporarily skipping the test on PyPy allows us to get tests passing again, until we can "
        "address this."
    ),
)
def test_devendoring_required():
    # type: () -> None
    # The cryptography distribution does not have a whl released for python3 on linux at version 2.5.
    # As a result, we're forced to build it under python3 and, prior to the fix for
    # https://github.com/pex-tool/pex/issues/661, this would fail using the vendored setuptools
    # inside pex.
    with temporary_dir() as td:
        cryptography_pex = os.path.join(td, "cryptography.pex")
        res = run_pex_command(["cryptography==2.5", "-o", cryptography_pex])
        res.assert_success()

        subprocess.check_call([cryptography_pex, "-c", "import cryptography"])



# Copyright 2014 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, print_function

import contextlib
import glob
import itertools
import os
import platform
import random
import re
import subprocess
import sys
from collections import Counter
from contextlib import contextmanager
from textwrap import dedent

from pex.atomic_directory import atomic_directory
from pex.common import open_zip, safe_mkdir, safe_mkdtemp, safe_rmtree, safe_sleep, temporary_dir
from pex.compatibility import to_unicode
from pex.dist_metadata import Distribution
from pex.enum import Enum
from pex.executor import Executor
from pex.interpreter import PythonInterpreter
from pex.pep_427 import install_wheel_chroot
from pex.pex import PEX
from pex.pex_builder import PEXBuilder
from pex.pex_info import PexInfo
from pex.pip.installation import get_pip
from pex.resolve.configured_resolver import ConfiguredResolver
from pex.resolve.resolver_configuration import PipConfiguration
from pex.typing import TYPE_CHECKING
from pex.util import named_temporary_file
from pex.venv.virtualenv import InstallationChoice, Virtualenv

try:
    from unittest import mock
except ImportError:
    import mock  # type: ignore[no-redef,import]

if TYPE_CHECKING:
    from typing import (
        Any,
        Callable,
        Dict,
        Iterable,
        Iterator,
        List,
        Mapping,
        Optional,
        Set,
        Text,
        Tuple,
        Union,
    )

    import attr  # vendor:skip
else:
    from pex.third_party import attr

PY_VER = sys.version_info[:2]
IS_PYPY = hasattr(sys, "pypy_version_info")
IS_PYPY2 = IS_PYPY and sys.version_info[0] == 2
IS_PYPY3 = IS_PYPY and sys.version_info[0] == 3
NOT_CPYTHON27 = IS_PYPY or PY_VER != (2, 7)
IS_LINUX = platform.system() == "Linux"
IS_MAC = platform.system() == "Darwin"
IS_X86_64 = platform.machine().lower() in ("amd64", "x86_64")
IS_ARM_64 = platform.machine().lower() in ("arm64", "aarch64")
IS_LINUX_X86_64 = IS_LINUX and IS_X86_64
IS_LINUX_ARM64 = IS_LINUX and IS_ARM_64
IS_MAC_X86_64 = IS_MAC and IS_X86_64
IS_MAC_ARM64 = IS_MAC and IS_ARM_64
NOT_CPYTHON27_OR_OSX = NOT_CPYTHON27 or not IS_LINUX


@contextlib.contextmanager
def temporary_filename():
    # type: () -> Iterator[str]
    """Creates a temporary filename.

    This is useful when you need to pass a filename to an API. Windows requires all handles to a
    file be closed before deleting/renaming it, so this makes it a bit simpler.
    """
    with named_temporary_file() as fp:
        fp.write(b"")
        fp.close()
        yield fp.name


def random_bytes(length):
    # type: (int) -> bytes
    return "".join(map(chr, (random.randint(ord("a"), ord("z")) for _ in range(length)))).encode(
        "utf-8"
    )


def get_dep_dist_names_from_pex(pex_path, match_prefix=""):
    # type: (str, str) -> Set[Text]
    """Given an on-disk pex, extract all of the unique first-level paths under `.deps`."""
    with open_zip(pex_path) as pex_zip:
        dep_gen = (f.split(os.sep)[1] for f in pex_zip.namelist() if f.startswith(".deps/"))
        return set(item for item in dep_gen if item.startswith(match_prefix))


@contextlib.contextmanager
def temporary_content(content_map, interp=None, seed=31337, perms=0o644):
    # type: (Mapping[str, Union[int, str]], Optional[Dict[str, Any]], int, int) -> Iterator[str]
    """Write content to disk where content is map from string => (int, string).

    If target is int, write int random bytes.  Otherwise write contents of string.
    """
    random.seed(seed)
    interp = interp or {}
    with temporary_dir() as td:
        for filename, size_or_content in content_map.items():
            dest = os.path.join(td, filename)
            safe_mkdir(os.path.dirname(dest))
            with open(dest, "wb") as fp:
                if isinstance(size_or_content, int):
                    fp.write(random_bytes(size_or_content))
                else:
                    fp.write((size_or_content % interp).encode("utf-8"))
            os.chmod(dest, perms)
        yield td


@contextlib.contextmanager
def make_project(
    name="my_project",  # type: str
    version="0.0.0",  # type: str
    zip_safe=True,  # type: bool
    install_reqs=None,  # type: Optional[List[str]]
    extras_require=None,  # type: Optional[Dict[str, List[str]]]
    entry_points=None,  # type: Optional[Union[str, Dict[str, List[str]]]]
    python_requires=None,  # type: Optional[str]
    universal=False,  # type: bool
    prepare_project=None,  # type: Optional[Callable[[str], None]]
):
    # type: (...) -> Iterator[str]
    project_content = {
        "setup.py": dedent(
            """
            from setuptools import setup
            
            setup(
            name=%(project_name)r,
            version=%(version)r,
            zip_safe=%(zip_safe)r,
            packages=[%(project_name)r],
            scripts=[
              'scripts/hello_world',
              'scripts/shell_script',
            ],
            package_data={%(project_name)r: ['package_data/*.dat']},
            install_requires=%(install_requires)r,
            extras_require=%(extras_require)r,
            entry_points=%(entry_points)r,
            python_requires=%(python_requires)r,
            options={'bdist_wheel': {'universal': %(universal)r}},
            )
            """
        ),
        "scripts/hello_world": '#!/usr/bin/env python\nprint("hello world from py script!")\n',
        "scripts/shell_script": "#!/usr/bin/env bash\necho hello world from shell script\n",
        os.path.join(name, "__init__.py"): 0,
        os.path.join(name, "my_module.py"): 'def do_something():\n  print("hello world!")\n',
        os.path.join(name, "package_data/resource1.dat"): 1000,
        os.path.join(name, "package_data/resource2.dat"): 1000,
    }  # type: Dict[str, Union[str, int]]

    interp = {
        "project_name": name,
        "version": version,
        "zip_safe": zip_safe,
        "install_requires": install_reqs or [],
        "extras_require": extras_require or {},
        "entry_points": entry_points or {},
        "python_requires": python_requires,
        "universal": universal,
    }

    with temporary_content(project_content, interp=interp) as td:
        if prepare_project:
            prepare_project(td)
        yield td


class WheelBuilder(object):
    """Create a wheel distribution from an unpacked setup.py-based project."""

    class BuildFailure(Exception):
        pass

    def __init__(
        self,
        source_dir,  # type: str
        interpreter=None,  # type: Optional[PythonInterpreter]
        wheel_dir=None,  # type: Optional[str]
        verify=True,  # type: bool
    ):
        # type: (...) -> None
        """Create a wheel from an unpacked source distribution in source_dir."""
        self._source_dir = source_dir
        self._wheel_dir = wheel_dir or safe_mkdtemp()
        self._interpreter = interpreter or PythonInterpreter.get()
        self._verify = verify

    def bdist(self):
        # type: () -> str
        get_pip(
            interpreter=self._interpreter,
            resolver=ConfiguredResolver(pip_configuration=PipConfiguration()),
        ).spawn_build_wheels(
            distributions=[self._source_dir],
            wheel_dir=self._wheel_dir,
            interpreter=self._interpreter,
            verify=self._verify,
        ).wait()
        dists = glob.glob(os.path.join(self._wheel_dir, "*.whl"))
        if len(dists) == 0:
            raise self.BuildFailure("No distributions were produced!")
        if len(dists) > 1:
            raise self.BuildFailure(
                "Ambiguous wheel distributions found: {dists}".format(dists=" ".join(dists))
            )
        return dists[0]


@contextlib.contextmanager
def built_wheel(
    name="my_project",  # type: str
    version="0.0.0",  # type: str
    zip_safe=True,  # type: bool
    install_reqs=None,  # type: Optional[List[str]]
    extras_require=None,  # type: Optional[Dict[str, List[str]]]
    entry_points=None,  # type: Optional[Union[str, Dict[str, List[str]]]]
    interpreter=None,  # type: Optional[PythonInterpreter]
    python_requires=None,  # type: Optional[str]
    universal=False,  # type: bool
    prepare_project=None,  # type: Optional[Callable[[str], None]]
    **kwargs  # type: Any
):
    # type: (...) -> Iterator[str]
    with make_project(
        name=name,
        version=version,
        zip_safe=zip_safe,
        install_reqs=install_reqs,
        extras_require=extras_require,
        entry_points=entry_points,
        python_requires=python_requires,
        universal=universal,
        prepare_project=prepare_project,
    ) as td:
        builder = WheelBuilder(td, interpreter=interpreter, **kwargs)
        yield builder.bdist()


@contextlib.contextmanager
def make_source_dir(
    name="my_project",  # type: str
    version="0.0.0",  # type: str
    install_reqs=None,  # type: Optional[List[str]]
    extras_require=None,  # type: Optional[Dict[str, List[str]]]
):
    # type: (...) -> Iterator[str]
    with make_project(
        name=name, version=version, install_reqs=install_reqs, extras_require=extras_require
    ) as td:
        yield td


@contextlib.contextmanager
def make_bdist(
    name="my_project",  # type: str
    version="0.0.0",  # type: str
    zip_safe=True,  # type: bool
    interpreter=None,  # type: Optional[PythonInterpreter]
    **kwargs  # type: Any
):
    # type: (...) -> Iterator[Distribution]
    with built_wheel(
        name=name, version=version, zip_safe=zip_safe, interpreter=interpreter, **kwargs
    ) as dist_location:
        yield install_wheel(dist_location)


def install_wheel(
    wheel,  # type: str
    interpreter=None,  # type: Optional[PythonInterpreter]
):
    # type: (...) -> Distribution
    install_dir = os.path.join(safe_mkdtemp(), os.path.basename(wheel))
    install_wheel_chroot(wheel_path=wheel, destination=install_dir)
    return Distribution.load(install_dir)


COVERAGE_PREAMBLE = """
try:
  from coverage import coverage
  cov = coverage(auto_data=True, data_suffix=True)
  cov.start()
except ImportError:
  pass
"""


def write_simple_pex(
    td,  # type: str
    exe_contents=None,  # type: Optional[str]
    dists=None,  # type: Optional[Iterable[Distribution]]
    sources=None,  # type: Optional[Iterable[Tuple[str, str]]]
    coverage=False,  # type: bool
    interpreter=None,  # type: Optional[PythonInterpreter]
    pex_info=None,  # type: Optional[PexInfo]
):
    # type: (...) -> PEXBuilder
    """Write a pex file that optionally contains an executable entry point.

    :param td: temporary directory path
    :param exe_contents: entry point python file
    :param dists: distributions to include, typically sdists or bdists
    :param sources: sources to include, as a list of pairs (env_filename, contents)
    :param coverage: include coverage header
    :param interpreter: a custom interpreter to use to build the pex
    :param pex_info: a custom PexInfo to use to build the pex.
    """
    dists = dists or []
    sources = sources or []

    safe_mkdir(td)

    pb = PEXBuilder(
        path=td,
        preamble=COVERAGE_PREAMBLE if coverage else None,
        interpreter=interpreter,
        pex_info=pex_info,
    )

    for dist in dists:
        pb.add_dist_location(dist.location if isinstance(dist, Distribution) else dist)

    for env_filename, contents in sources:
        src_path = os.path.join(td, env_filename)
        safe_mkdir(os.path.dirname(src_path))
        with open(src_path, "w") as fp:
            fp.write(contents)
        pb.add_source(src_path, env_filename)

    if exe_contents:
        with open(os.path.join(td, "exe.py"), "w") as fp:
            fp.write(exe_contents)
        pb.set_executable(os.path.join(td, "exe.py"))

    pb.freeze()

    return pb


def re_exact(text):
    # type: (str) -> str
    return r"^{escaped}$".format(escaped=re.escape(text))


@attr.s(frozen=True)
class IntegResults(object):
    """Convenience object to return integration run results."""

    output = attr.ib()  # type: Text
    error = attr.ib()  # type: Text
    return_code = attr.ib()  # type: int

    def assert_success(
        self,
        expected_output_re=None,  # type: Optional[str]
        expected_error_re=None,  # type: Optional[str]
        re_flags=0,  # type: int
    ):
        # type: (...) -> None
        assert (
            self.return_code == 0
        ), "integration test failed: return_code={}, output={}, error={}".format(
            self.return_code, self.output, self.error
        )
        self.assert_output(expected_output_re, expected_error_re, re_flags)

    def assert_failure(
        self,
        expected_error_re=None,  # type: Optional[str]
        expected_output_re=None,  # type: Optional[str]
        re_flags=0,  # type: int
    ):
        # type: (...) -> None
        assert self.return_code != 0
        self.assert_output(expected_output_re, expected_error_re, re_flags)

    def assert_output(
        self,
        expected_output_re=None,  # type: Optional[str]
        expected_error_re=None,  # type: Optional[str]
        re_flags=0,  # type: int
    ):
        if expected_output_re:
            assert re.match(
                expected_output_re, self.output, flags=re_flags
            ), "Failed to match re: {re!r} against:\n{output}".format(
                re=expected_output_re, output=self.output
            )
        if expected_error_re:
            assert re.match(
                expected_error_re, self.error, flags=re_flags
            ), "Failed to match re: {re!r} against:\n{output}".format(
                re=expected_error_re, output=self.error
            )


def create_pex_command(
    args=None,  # type: Optional[Iterable[str]]
    python=None,  # type: Optional[str]
    quiet=False,  # type: bool
):
    # type: (...) -> List[str]
    cmd = [python or sys.executable, "-mpex"]
    if not quiet:
        cmd.append("-vvvvv")
    if args:
        cmd.extend(args)
    return cmd


def run_pex_command(
    args,  # type: Iterable[str]
    env=None,  # type: Optional[Dict[str, str]]
    python=None,  # type: Optional[str]
    quiet=False,  # type: bool
    cwd=None,  # type: Optional[str]
):
    # type: (...) -> IntegResults
    """Simulate running pex command for integration testing.

    This is different from run_simple_pex in that it calls the pex command rather than running a
    generated pex.  This is useful for testing end to end runs with specific command line arguments
    or env options.
    """
    cmd = create_pex_command(args, python=python, quiet=quiet)
    process = Executor.open_process(
        cmd=cmd, env=env, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    output, error = process.communicate()
    return IntegResults(output.decode("utf-8"), error.decode("utf-8"), process.returncode)


def run_simple_pex(
    pex,  # type: str
    args=(),  # type: Iterable[str]
    interpreter=None,  # type: Optional[PythonInterpreter]
    stdin=None,  # type: Optional[bytes]
    **kwargs  # type: Any
):
    # type: (...) -> Tuple[bytes, int]
    p = PEX(pex, interpreter=interpreter)
    process = p.run(
        args=args,
        blocking=False,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        **kwargs
    )
    stdout, _ = process.communicate(input=stdin)
    return stdout.replace(b"\r", b""), process.returncode


def run_simple_pex_test(
    body,  # type: str
    args=(),  # type: Iterable[str]
    env=None,  # type: Optional[Mapping[str, str]]
    dists=None,  # type: Optional[Iterable[Distribution]]
    coverage=False,  # type: bool
    interpreter=None,  # type: Optional[PythonInterpreter]
):
    # type: (...) -> Tuple[bytes, int]
    with temporary_dir() as td1, temporary_dir() as td2:
        pb = write_simple_pex(td1, body, dists=dists, coverage=coverage, interpreter=interpreter)
        pex = os.path.join(td2, "app.pex")
        pb.build(pex)
        return run_simple_pex(pex, args=args, env=env, interpreter=interpreter)


def bootstrap_python_installer(dest):
    # type: (str) -> None
    for _ in range(3):
        try:
            subprocess.check_call(["git", "clone", "https://github.com/pyenv/pyenv.git", dest])
        except subprocess.CalledProcessError as e:
            print("caught exception: %r" % e)
            continue
        else:
            break
    else:
        raise RuntimeError("Helper method could not clone pyenv from git after 3 tries")


# NB: We keep the pool of bootstrapped interpreters as small as possible to avoid timeouts in CI
# otherwise encountered when fetching and building too many on a cache miss. In the past we had
# issues with the combination of 7 total unique interpreter versions and a Travis-CI timeout of 50
# minutes for a shard.
# N.B.: Make sure to stick to versions that have binary releases for all supported platforms to
# support use of pyenv-win which does not build from source, just running released installers
# robotically instead.
PY27 = "2.7.18"
PY38 = "3.8.10"
PY39 = "3.9.13"
PY310 = "3.10.7"

ALL_PY_VERSIONS = (PY27, PY38, PY39, PY310)
_ALL_PY_VERSIONS_TO_VERSION_INFO = {
    version: tuple(map(int, version.split("."))) for version in ALL_PY_VERSIONS
}


PEX_TEST_DEV_ROOT = os.path.abspath(
    os.path.expanduser(os.environ.get("_PEX_TEST_DEV_ROOT", "~/.pex_dev"))
)


def ensure_python_distribution(version):
    # type: (str) -> Tuple[str, str, str, Callable[[Iterable[str]], Text]]
    if version not in ALL_PY_VERSIONS:
        raise ValueError("Please constrain version to one of {}".format(ALL_PY_VERSIONS))

    pyenv_root = os.path.join(PEX_TEST_DEV_ROOT, "pyenv")
    interpreter_location = os.path.join(pyenv_root, "versions", version)

    pyenv = os.path.join(pyenv_root, "bin", "pyenv")
    pyenv_env = os.environ.copy()
    pyenv_env["PYENV_ROOT"] = pyenv_root

    pip = os.path.join(interpreter_location, "bin", "pip")

    with atomic_directory(target_dir=pyenv_root) as target_dir:
        if not target_dir.is_finalized():
            bootstrap_python_installer(target_dir.work_dir)

    with atomic_directory(target_dir=interpreter_location) as interpreter_target_dir:
        if not interpreter_target_dir.is_finalized():
            subprocess.check_call(
                [
                    "git",
                    "--git-dir={}".format(os.path.join(pyenv_root, ".git")),
                    "--work-tree={}".format(pyenv_root),
                    "reset",
                    "--hard",
                    "v2.3.8",
                ]
            )
            env = pyenv_env.copy()
            if sys.platform.lower().startswith("linux"):
                env["CONFIGURE_OPTS"] = "--enable-shared"
                # The pyenv builder detects `--enable-shared` and sets up `RPATH` via
                # `LDFLAGS=-Wl,-rpath=... $LDFLAGS` to ensure the built python binary links the
                # correct libpython shared lib. Some versions of compiler set the `RUNPATH` instead
                # though which is searched _after_ the `LD_LIBRARY_PATH` environment variable. To
                # ensure an inopportune `LD_LIBRARY_PATH` doesn't fool the pyenv python binary into
                # linking the wrong libpython, force `RPATH`, which is searched 1st by the linker,
                # with with `--disable-new-dtags`.
                env["LDFLAGS"] = "-Wl,--disable-new-dtags"
            subprocess.check_call([pyenv, "install", "--keep", version], env=env)
            subprocess.check_call([pip, "install", "-U", "pip<22.1"])

    major, minor = version.split(".")[:2]
    python = os.path.join(
        interpreter_location, "bin", "python{major}.{minor}".format(major=major, minor=minor)
    )

    def run_pyenv(args):
        # type: (Iterable[str]) -> Text
        return to_unicode(subprocess.check_output([pyenv] + list(args), env=pyenv_env))

    return interpreter_location, python, pip, run_pyenv


def ensure_python_venv(
    version,  # type: str
    latest_pip=True,  # type: bool
    system_site_packages=False,  # type: bool
):
    # type: (...) -> Tuple[str, str]
    _, python, pip, _ = ensure_python_distribution(version)
    venv = safe_mkdtemp()
    if _ALL_PY_VERSIONS_TO_VERSION_INFO[version][0] == 3:
        args = [python, "-m", "venv", venv]
        if system_site_packages:
            args.append("--system-site-packages")
        subprocess.check_call(args=args)
    else:
        subprocess.check_call(args=[pip, "install", "virtualenv==16.7.10"])
        args = [python, "-m", "virtualenv", venv, "-q"]
        if system_site_packages:
            args.append("--system-site-packages")
        subprocess.check_call(args=args)
    python, pip = tuple(os.path.join(venv, "bin", exe) for exe in ("python", "pip"))
    if latest_pip:
        subprocess.check_call(args=[pip, "install", "-U", "pip<22.1"])
    return python, pip


def ensure_python_interpreter(version):
    # type: (str) -> str
    _, python, _, _ = ensure_python_distribution(version)
    return python


class InterpreterImplementation(Enum["InterpreterImplementation.Value"]):
    class Value(Enum.Value):
        pass

    CPython = Value("CPython")
    PyPy = Value("PyPy")


def find_python_interpreter(
    version=(),  # type: Tuple[int, ...]
    implementation=InterpreterImplementation.CPython,  # type: InterpreterImplementation.Value
):
    # type: (...) -> Optional[str]
    for pyenv_version, penv_version_info in _ALL_PY_VERSIONS_TO_VERSION_INFO.items():
        if version and version == penv_version_info[: len(version)]:
            return ensure_python_interpreter(pyenv_version)

    for interpreter in PythonInterpreter.iter():
        if version != interpreter.version[: len(version)]:
            continue
        if implementation != InterpreterImplementation.for_value(interpreter.identity.interpreter):
            continue
        return interpreter.binary

    return None


def python_venv(
    python,  # type: str
    system_site_packages=False,  # type: bool
    venv_dir=None,  # type: Optional[str]
):
    # type: (...) -> Tuple[str, str]
    venv = Virtualenv.create(
        venv_dir=venv_dir or safe_mkdtemp(),
        interpreter=PythonInterpreter.from_binary(python),
        system_site_packages=system_site_packages,
        install_pip=InstallationChoice.YES,
    )
    return venv.interpreter.binary, venv.bin_path("pip")


def all_pythons():
    # type: () -> Tuple[str, ...]
    return tuple(ensure_python_interpreter(version) for version in ALL_PY_VERSIONS)


@attr.s(frozen=True)
class VenvFactory(object):
    python_version = attr.ib()  # type: str
    _factory = attr.ib()  # type: Callable[[], Tuple[str, str]]

    def create_venv(self):
        # type: () -> Tuple[str, str]
        return self._factory()


def all_python_venvs(system_site_packages=False):
    # type: (bool) -> Iterable[VenvFactory]
    return tuple(
        VenvFactory(
            python_version=version,
            factory=lambda: ensure_python_venv(version, system_site_packages=system_site_packages),
        )
        for version in ALL_PY_VERSIONS
    )


@contextmanager
def environment_as(**kwargs):
    # type: (**Any) -> Iterator[None]
    existing = {key: os.environ.get(key) for key in kwargs}

    def adjust_environment(mapping):
        for key, value in mapping.items():
            if value is not None:
                os.environ[key] = str(value)
            else:
                os.environ.pop(key, None)

    adjust_environment(kwargs)
    try:
        yield
    finally:
        adjust_environment(existing)


@contextmanager
def pushd(directory):
    # type: (Text) -> Iterator[None]
    cwd = os.getcwd()
    try:
        os.chdir(directory)
        yield
    finally:
        os.chdir(cwd)


def make_env(**kwargs):
    # type: (**Any) -> Dict[str, str]
    """Create a copy of the current environment with the given modifications.

    The given kwargs add to or update the environment when they have a non-`None` value. When they
    have a `None` value, the environment variable is removed from the environment.

    All non-`None` values are converted to strings by apply `str`.
    """
    env = os.environ.copy()
    env.update((k, str(v)) for k, v in kwargs.items() if v is not None)
    for k, v in kwargs.items():
        if v is None:
            env.pop(k, None)
    return env


def run_commands_with_jitter(
    commands,  # type: Iterable[Iterable[str]]
    path_argument,  # type: str
    extra_env=None,  # type: Optional[Mapping[str, str]]
    delay=2.0,  # type: float
):
    # type: (...) -> List[str]
    """Runs the commands with tactics that attempt to introduce randomness in outputs.

    Each command will run against a clean Pex cache with a unique path injected as the value for
    `path_argument`. A unique `PYTHONHASHSEED` is set in the environment for each execution as well.

    Additionally, a delay is inserted between executions. By default, this delay is 2s to ensure zip
    precision is stressed. See: https://pkware.cachefly.net/webdocs/casestudies/APPNOTE.TXT.
    """
    td = safe_mkdtemp()
    pex_root = os.path.join(td, "pex_root")

    paths = []
    for index, command in enumerate(commands):
        path = os.path.join(td, str(index))
        cmd = list(command) + [path_argument, path]

        # Note that we change the `PYTHONHASHSEED` to ensure that there are no issues resulting
        # from the random seed, such as data structures, as Tox sets this value by default.
        # See:
        # https://tox.readthedocs.io/en/latest/example/basic.html#special-handling-of-pythonhashseed
        env = make_env(PEX_ROOT=pex_root, PYTHONHASHSEED=(index * 497) + 4)
        if extra_env:
            env.update(extra_env)

        if index > 0:
            safe_sleep(delay)

        # Ensure the PEX is fully rebuilt.
        safe_rmtree(pex_root)
        subprocess.check_call(args=cmd, env=env)
        paths.append(path)
    return paths


def run_command_with_jitter(
    args,  # type: Iterable[str]
    path_argument,  # type: str
    extra_env=None,  # type: Optional[Mapping[str, str]]
    delay=2.0,  # type: float
    count=3,  # type: int
):
    # type: (...) -> List[str]
    """Runs the command `count` times in an attempt to introduce randomness.

    Each run of the command will run against a clean Pex cache with a unique path injected as the
    value for `path_argument`. A unique `PYTHONHASHSEED` is set in the environment for each
    execution as well.

    Additionally, a delay is inserted between executions. By default, this delay is 2s to ensure zip
    precision is stressed. See: https://pkware.cachefly.net/webdocs/casestudies/APPNOTE.TXT.
    """
    return run_commands_with_jitter(
        commands=list(itertools.repeat(list(args), count)),
        path_argument=path_argument,
        extra_env=extra_env,
        delay=delay,
    )


def pex_project_dir():
    # type: () -> str
    try:
        return os.environ["_PEX_TEST_PROJECT_DIR"]
    except KeyError:
        sys.exit("Pex tests must be run via tox.")


class NonDeterministicWalk:
    """A wrapper around `os.walk` that makes it non-deterministic.

    Makes sure that directories and files are always returned in a different
    order each time it is called.

    Typically used like: `unittest.mock.patch("os.walk", new=NonDeterministicWalk())`
    """

    def __init__(self):
        self._counter = Counter()  # type: Counter[str, int]
        self._original_walk = os.walk

    def __call__(self, *args, **kwargs):
        # type: (*Any, **Any) -> Iterator[Tuple[str, List[str], List[str]]]
        for root, dirs, files in self._original_walk(*args, **kwargs):
            self._increment_counter(root)
            dirs[:] = self._rotate(root, dirs)
            files[:] = self._rotate(root, files)
            yield root, dirs, files

    def _increment_counter(self, counter_key):
        # type: (str) -> int
        self._counter[counter_key] += 1
        return self._counter[counter_key]

    def _rotate(self, counter_key, x):
        # type: (str, List[str]) -> List[str]
        if not x:
            return x
        rotate_by = self._counter[counter_key] % len(x)
        return x[-rotate_by:] + x[:-rotate_by]



# Copyright 2023 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import stat
import subprocess

from pex.common import safe_rmtree
from pex.typing import TYPE_CHECKING
from pex.venv.virtualenv import InstallationChoice, Virtualenv

if TYPE_CHECKING:
    from typing import Any


def test_read_only_venv(
    tmpdir,  # type: Any
    pex_project_dir,  # type: str
):
    # type: (...) -> None

    venv_dir = os.path.join(str(tmpdir), "venv")
    venv = Virtualenv.create(venv_dir, install_pip=InstallationChoice.UPGRADED)
    subprocess.check_call(args=[venv.bin_path("pip"), "install", pex_project_dir])

    pex_root = os.path.join(str(tmpdir), "pex_root")

    def assert_pex_works():
        safe_rmtree(pex_root)
        assert (
            "Moo!"
            in subprocess.check_output(
                args=[
                    venv.bin_path("pex"),
                    "--pex-root",
                    pex_root,
                    "--runtime-pex-root",
                    pex_root,
                    "cowsay==5.0",
                    "-c",
                    "cowsay",
                    "--",
                    "Moo!",
                ]
            ).decode("utf-8")
        )

    assert_pex_works()

    write_mask = stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH
    for root, dirs, files in os.walk(venv.site_packages_dir, topdown=False):
        for path in files + dirs:
            abs_path = os.path.join(root, path)
            os.chmod(abs_path, os.stat(abs_path).st_mode & ~write_mask)

    assert_pex_works()



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os
import re
from collections import OrderedDict

from pex.common import pluralize
from pex.compatibility import string
from pex.dist_metadata import Requirement, RequirementParseError
from pex.third_party.packaging.specifiers import InvalidSpecifier, SpecifierSet
from pex.typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import Any, List, Mapping, Tuple

    import attr  # vendor:skip
    import toml  # vendor:skip
else:
    from pex.third_party import attr, toml


_UNSPECIFIED_SOURCE = "<unspecified source>"
_SPEC_URL = (
    "https://packaging.python.org/specifications/inline-script-metadata#inline-script-metadata"
)


class InvalidMetadataError(ValueError):
    """Indicates invalid PEP-723 script metadata."""


@attr.s(frozen=True)
class MetadataBlock(object):
    source = attr.ib()  # type: str
    start_line = attr.ib()  # type: int
    type = attr.ib()  # type: str
    content = attr.ib()  # type: Tuple[str, ...]

    def create_metadata_error(self, problem_clause):
        # type: (str) -> InvalidMetadataError
        return InvalidMetadataError(
            "The script metadata found in {source} starting at line {line} {problem_clause}.\n"
            "See: {spec_url}".format(
                source=self.source,
                line=self.start_line,
                problem_clause=problem_clause,
                spec_url=_SPEC_URL,
            )
        )

    def parse_metadata(self):
        # type: () -> Mapping[str, Any]
        stripped_content = os.linesep.join(
            line[2:] if line.startswith("# ") else line[1:] for line in self.content
        )
        try:
            return cast("Mapping[str, Any]", toml.loads(stripped_content))
        except toml.TomlDecodeError as e:
            raise self.create_metadata_error("embeds malformed toml: {err}".format(err=e))


@attr.s
class ParseState(object):
    start_type = attr.ib(default="", init=False)  # type: str
    start_line = attr.ib(default=0, init=False)  # type: int
    end_line = attr.ib(default=0, init=False)  # type: int

    def reset(
        self,
        start_type="",  # type: str
        start_line=0,  # type: int
    ):
        # type: (...) -> None
        self.start_type = start_type
        self.start_line = start_line
        self.end_line = 0

    @property
    def started(self):
        # type: () -> bool
        return bool(self.start_type) and self.start_line > 0

    @property
    def finished(self):
        # type: () -> bool
        return self.started and self.end_line >= self.start_line


def parse_metadata_blocks(
    script,  # type: str
    source=_UNSPECIFIED_SOURCE,  # type: str
):
    # type: (...) -> Mapping[str, MetadataBlock]

    lines = script.splitlines()
    metadata_blocks = OrderedDict()  # type: OrderedDict[str, List[MetadataBlock]]
    parse_state = ParseState()

    def add_metadata_block():
        # type: () -> None
        metadata_blocks.setdefault(parse_state.start_type, []).append(
            MetadataBlock(
                source=source,
                start_line=parse_state.start_line,
                type=parse_state.start_type,
                content=tuple(lines[parse_state.start_line : parse_state.end_line - 1]),
            )
        )
        parse_state.reset()

    for line_no, line in enumerate(lines, start=1):
        start = re.match(r"^# /// (?P<type>[a-zA-Z0-9-]+)$", line)
        if start and parse_state.started and not parse_state.finished:
            raise InvalidMetadataError(
                "The script metadata found in {source} contains a `# /// {outer_type}` block "
                "beginning on line {outer_start_line} that is followed by a `# /// {inner_type}` "
                "block beginning on line {inner_start_line} before the {outer_type!r} block "
                "starting on line {outer_start_line} is closed.\n"
                "Metadata blocks must be closed before a new metadata block can begin.\n"
                "See: {spec_url}".format(
                    source=source,
                    outer_type=parse_state.start_type,
                    outer_start_line=parse_state.start_line,
                    inner_type=start.group("type"),
                    inner_start_line=line_no,
                    spec_url=_SPEC_URL,
                )
            )
        elif start:
            if parse_state.finished:
                add_metadata_block()
            parse_state.reset(start_type=start.group("type"), start_line=line_no)
        elif parse_state.started and "# ///" == line:
            parse_state.end_line = line_no
        elif line != "#" and not line.startswith("# "):
            if parse_state.finished and parse_state.end_line == line_no - 1:
                add_metadata_block()
            else:
                parse_state.reset()
    if parse_state.finished:
        add_metadata_block()

    over_abundant_blocks = []  # type: List[str]
    for type_, blocks in metadata_blocks.items():
        count = len(blocks)
        if count > 1:
            over_abundant_blocks.append(
                "+ {count} `# /// {type}` metadata blocks beginning on lines {lines}.".format(
                    count=count,
                    type=type_,
                    lines="{lines} and {last_line}".format(
                        lines=", ".join(map(str, (block.start_line for block in blocks[:-1]))),
                        last_line=blocks[-1].start_line,
                    ),
                )
            )

    if over_abundant_blocks:
        raise InvalidMetadataError(
            "Found {count} metadata block {types} in {source} with more than one appearance:\n"
            "{over_abundant_blocks}\n"
            "At most one metadata block of each type is allowed.\n"
            "See: {spec_url}".format(
                count=len(over_abundant_blocks),
                types=pluralize(over_abundant_blocks, "type"),
                source=source,
                over_abundant_blocks="\n".join(over_abundant_blocks),
                spec_url=_SPEC_URL,
            )
        )
    return {type_: blocks[0] for type_, blocks in metadata_blocks.items()}


@attr.s(frozen=True)
class ScriptMetadata(object):
    @classmethod
    def parse(
        cls,
        script,  # type: str
        source=_UNSPECIFIED_SOURCE,  # type: str
    ):
        # type: (...) -> ScriptMetadata

        # The spec this code follows was defined in PEP-723: https://peps.python.org/pep-0723/
        # and now lives here:
        # https://packaging.python.org/specifications/inline-script-metadata#inline-script-metadata
        script_metadata_block = parse_metadata_blocks(script, source=source).get("script")
        if not script_metadata_block:
            return cls()
        script_metadata = script_metadata_block.parse_metadata()

        raw_dependencies = script_metadata.get("dependencies", [])
        if not isinstance(raw_dependencies, list):
            raise script_metadata_block.create_metadata_error(
                "contains an invalid `dependencies` value of type `{type}`\n"
                "Expected a list of dependency specifier strings".format(
                    type=type(raw_dependencies).__name__
                )
            )

        invalid_dependencies = []  # type: List[str]
        dependencies = []  # type: List[Requirement]
        for index, req in enumerate(raw_dependencies):
            try:
                dependencies.append(Requirement.parse(req))
            except RequirementParseError as e:
                invalid_dependencies.append(
                    "+ dependencies[{index}] {req!r}: {err}".format(index=index, req=req, err=e)
                )
        if invalid_dependencies:
            raise script_metadata_block.create_metadata_error(
                "contains a `dependencies` list with {count} invalid dependency {specifiers}:\n"
                "{invalid_dependencies}".format(
                    count=len(invalid_dependencies),
                    specifiers=pluralize(invalid_dependencies, "specifier"),
                    invalid_dependencies="\n".join(invalid_dependencies),
                )
            )

        raw_requires_python = script_metadata.get("requires-python", "")
        if not isinstance(raw_requires_python, string):
            raise script_metadata_block.create_metadata_error(
                "contains an invalid `requires-python` value of type `{type}`\n"
                "Expected a version specifier string".format(
                    type=type(raw_requires_python).__name__
                )
            )
        try:
            requires_python = SpecifierSet(raw_requires_python)
        except InvalidSpecifier as e:
            raise script_metadata_block.create_metadata_error(
                "contains an invalid `requires-python` value {value!r}: {err}".format(
                    value=raw_requires_python, err=e
                )
            )

        return cls(dependencies=tuple(dependencies), requires_python=requires_python, source=source)

    dependencies = attr.ib(default=())  # type: Tuple[Requirement, ...]
    requires_python = attr.ib(default=SpecifierSet())  # type: SpecifierSet
    source = attr.ib(default=_UNSPECIFIED_SOURCE)  # type: str

    def __bool__(self):
        # type: () -> bool
        return bool(self.dependencies) or bool(self.requires_python)

    # N.B.: For Python 2.7.
    __nonzero__ = __bool__



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import subprocess
import sys

from pex.interpreter import PythonInterpreter
from pex.typing import TYPE_CHECKING
from testing import PY310, ensure_python_interpreter, run_pex_command
from testing.cli import run_pex3

if TYPE_CHECKING:
    from typing import Any


def test_lock_use_no_build_wheel(tmpdir):
    # type: (Any)-> None

    lock = os.path.join(str(tmpdir), "black.lock")
    run_pex3(
        "lock",
        "create",
        "black==22.8.0",
        "-o",
        lock,
        "--style",
        "universal",
        "--indent",
        "2",
        "--interpreter-constraint",
        "CPython==3.10.*",
        "--wheel",
        "--no-build",
    ).assert_success()

    pex = os.path.join(str(tmpdir), "black.pex")
    python = sys.executable if sys.version_info[:2] == (3, 10) else ensure_python_interpreter(PY310)
    run_pex_command(
        args=[
            "-o",
            pex,
            "--python",
            python,
            "-c",
            "black",
            "--lock",
            lock,
            "--wheel",
            "--no-build",
        ]
    ).assert_success()

    output = subprocess.check_output(args=[pex, "--version"])
    assert (
        "black.pex, 22.8.0 (compiled: {compiled})".format(
            compiled="no" if PythonInterpreter.from_binary(python).is_pypy else "yes"
        )
        in output.decode("utf-8").splitlines()
    )



# Copyright 2023 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import sys

import pytest

from pex.typing import TYPE_CHECKING
from testing.cli import run_pex3

if TYPE_CHECKING:
    from typing import Any


@pytest.mark.skipif(
    sys.version_info[:2] < (3, 5) or sys.version_info[:2] >= (3, 12),
    reason=(
        "The tested distribution is only compatible with Python >= 3.5 and it requires lxml (4.9.2)"
        " which only has pre-built wheels available through 3.11."
    ),
)
def test_build_system_no_build_backend(tmpdir):
    # type: (Any) -> None

    run_pex3("lock", "create", "xmlsec==1.3.13").assert_success()



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import sys
from datetime import datetime
from pathlib import PurePath

sys.path.insert(0, str(PurePath(__file__).parent.parent))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

# Note: must come after the sys.path manipulation above.
from pex.version import __version__ as PEX_VERSION  # isort:skip

project = "pex"
version = ".".join(PEX_VERSION.split(".")[:2])
release = PEX_VERSION
copyright = f"{datetime.now().year}, Pex project contributors"
author = "Pex project contributors"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

# Note: the vars extension is housed in _ext.
sys.path.insert(0, str(PurePath(__file__).parent / "_ext"))
extensions = [
    "myst_parser",
    "sphinx_pex",
]

source_suffix = {
    ".md": "markdown",
    ".rst": "restructuredtext",
}

suppress_warnings = [
    # Otherwise epub warns (and we treat warinings as errors) when it finds .doctrees/ files, which it should not
    # consider anyhow.
    "epub.unknown_project_files"
]

templates_path = [
    "_templates",
]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output
# https://myst-parser.readthedocs.io/en/latest/configuration.html
# https://pradyunsg.me/furo/customisation/

myst_enable_extensions = [
    "linkify",
]

import sphinx_pex
from sphinx_pex import SVGIcon

html_title = f"Pex Docs (v{release})"
html_theme = "furo"
html_favicon = "_static/pex-icon.png"
html_static_path = sphinx_pex.html_static_path()


html_theme_options = {
    "light_logo": "pex-logo-light.png",
    "dark_logo": "pex-logo-dark.png",
    "sidebar_hide_name": True,
    "source_repository": "https://github.com/pex-tool/pex/",
    "source_branch": "main",
    "source_directory": "docs/",
    "footer_icons": [
        icon.as_furo_footer_icon()
        for icon in [
            SVGIcon.load_if_static_asset_exists(
                name="PDF", icon_asset=PurePath("pdf.svg"), static_asset=PurePath("pex.pdf")
            ),
            SVGIcon.load(
                name="PyPI",
                icon_asset=PurePath("python.svg"),
                url=f"https://pypi.org/project/pex/{PEX_VERSION}/",
            ),
            SVGIcon.load(
                name="Download",
                icon_asset=PurePath("download.svg"),
                url=f"https://github.com/pex-tool/pex/releases/download/v{PEX_VERSION}/pex",
            ),
            SVGIcon.load(
                name="Source",
                icon_asset=PurePath("github.svg"),
                url="https://github.com/pex-tool/pex",
            ),
        ]
        if icon
    ],
}

# -- Options for Sphinx-SimplePDF output -------------------------------------------------
# https://sphinx-simplepdf.readthedocs.io/en/latest/configuration.html

simplepdf_vars = {
    "cover": "black",
    # Confusing! The white gets applied to the page number color (which appears inside a rusty red
    # box) and the back cover text color. Neither white nor black work well for the back page text
    # color since the text intersects the thick black bottom arc of the Pex P-egg. The yolk color
    # works well enough in both these spots though.
    "white": "#ffee00",
    "cover-bg": "url(pex-icon.png) no-repeat center",
}



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import glob
import os.path
import subprocess
from textwrap import dedent

import pytest

from pex.common import is_exe, safe_open
from pex.pep_503 import ProjectName
from pex.typing import TYPE_CHECKING
from pex.venv.virtualenv import InstallationChoice, Virtualenv

if TYPE_CHECKING:
    from typing import Any, List


def create_installation_choice_params(project_name):
    # type: (str) -> List[Any]
    return [
        pytest.param(
            installation_choice,
            id="{project_name}:{installation_choice}".format(
                project_name=project_name, installation_choice=installation_choice
            ),
        )
        for installation_choice in InstallationChoice.values()
    ]


@pytest.mark.parametrize("install_pip", create_installation_choice_params("pip"))
@pytest.mark.parametrize("install_setuptools", create_installation_choice_params("setuptools"))
@pytest.mark.parametrize("install_wheel", create_installation_choice_params("wheel"))
def test_venv_with_installs(
    tmpdir,  # type: Any
    install_pip,  # type: InstallationChoice.Value
    install_setuptools,  # type: InstallationChoice.Value
    install_wheel,  # type: InstallationChoice.Value
):
    # type: (...) -> None

    if install_pip is InstallationChoice.NO and (
        (install_setuptools is not InstallationChoice.NO)
        or (install_wheel is not InstallationChoice.NO)
    ):
        with pytest.raises(ValueError, match=r"^Installation of Pip is required in order to "):
            Virtualenv.create(
                str(tmpdir),
                install_pip=install_pip,
                install_setuptools=install_setuptools,
                install_wheel=install_wheel,
            )
        return

    venv = Virtualenv.create(
        str(tmpdir),
        install_pip=install_pip,
        install_setuptools=install_setuptools,
        install_wheel=install_wheel,
    )
    expected_projects = []
    if install_pip is not InstallationChoice.NO:
        expected_projects.append(ProjectName("pip"))
        # Pip gets installed with setuptools for Python < 3.12.
        if venv.interpreter.version[:2] < (3, 12):
            expected_projects.append(ProjectName("setuptools"))
    if install_setuptools is not InstallationChoice.NO:
        expected_projects.append(ProjectName("setuptools"))
    if install_wheel is not InstallationChoice.NO:
        expected_projects.append(ProjectName("wheel"))
    assert frozenset(expected_projects) == frozenset(
        dist.metadata.project_name for dist in venv.iter_distributions(rescan=True)
    )


def test_bdist_pex_under_tox(
    tmpdir,  # type: Any
    pex_project_dir,  # type: str
):
    # type: (...) -> None

    vend_dir = os.path.join(str(tmpdir), "venv")
    venv = Virtualenv.create(venv_dir=vend_dir, install_pip=InstallationChoice.YES)
    venv.interpreter.execute(args=["-m", "pip", "install", "tox"])
    tox = venv.bin_path("tox")
    assert is_exe(tox)

    project_dir = os.path.join(str(tmpdir), "project")
    with safe_open(os.path.join(project_dir, "entry.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                import cowsay


                def main():
                    print(cowsay.__version__)
                """
            )
        )
    with safe_open(os.path.join(project_dir, "setup.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                import setuptools

                setuptools.setup(
                    name="repro",
                    version="0.0.1",
                    install_requires=["cowsay==5.0"],
                    entry_points={
                        "console_scripts": [
                            "entry = entry:main",
                        ],
                    },
                    py_modules=["entry"],
                )
                """
            )
        )
    with safe_open(os.path.join(project_dir, "tox.ini"), "w") as fp:
        fp.write(
            dedent(
                """\
                [testenv:bundle]
                deps =
                    setuptools
                    {pex}
                commands =
                    {{envpython}} setup.py bdist_pex \
                        --bdist-dir=dist --pex-args=--disable-cache --bdist-all
                """
            ).format(pex=pex_project_dir)
        )

    subprocess.check_call(args=[tox, "-e", "bundle"], cwd=project_dir)
    pexes = glob.glob(os.path.join(project_dir, "dist", "*"))
    assert 1 == len(pexes)

    pex = pexes[0]
    assert b"5.0\n" == subprocess.check_output(args=[pex])



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os
import subprocess
from textwrap import dedent

import pytest

from pex.common import chmod_plus_x, is_exe, safe_mkdtemp
from pex.typing import TYPE_CHECKING
from testing import pex_project_dir

if TYPE_CHECKING:
    from typing import Optional

    import attr  # vendor:skip
else:
    from pex.third_party import attr


HAVE_DOCKER = any(
    is_exe(os.path.join(entry, "docker"))
    for entry in os.environ.get("PATH", os.path.defpath).split(os.pathsep)
)

skip_unless_docker = pytest.mark.skipif(not HAVE_DOCKER, reason="This test needs docker to run.")


@attr.s(frozen=True)
class DockerVirtualenvRunner(object):
    """Runs code in a venv created by Virtualenv at /virtualenv.venv in a docker container."""

    @classmethod
    def create(
        cls,
        base_image,  # type: str
        python="python",  # type: str
        virtualenv_version=None,  # type: Optional[str]
        tmpdir=None,  # type: Optional[str]
    ):
        # type: (...) -> DockerVirtualenvRunner

        test_script = os.path.join(tmpdir or safe_mkdtemp(), "test.sh")
        with open(test_script, "w") as fp:
            fp.write(
                dedent(
                    """\
                    #!/usr/bin/env bash

                    set -euo pipefail

                    {python} -mvenv /setup.venv >&2
                    /setup.venv/bin/pip install {virtualenv_requirement} >&2
                    /setup.venv/bin/virtualenv /virtualenv.venv >&2
                    PYTHONPATH=/code /virtualenv.venv/bin/python "$@"
                    """
                ).format(
                    python=python,
                    virtualenv_requirement=(
                        "virtualenv=={virtualenv_version}".format(
                            virtualenv_version=virtualenv_version
                        )
                        if virtualenv_version
                        else "virtualenv"
                    ),
                )
            )
        chmod_plus_x(test_script)

        return cls(base_image=base_image, test_script=test_script)

    base_image = attr.ib()  # type: str
    test_script = attr.ib()  # type: str

    def run(
        self,
        python_code,  # type: str
        *args  # type: str
    ):
        # type: (...) -> bytes

        if not HAVE_DOCKER:
            pytest.skip("This test needs docker to run.")

        return subprocess.check_output(
            args=[
                "docker",
                "run",
                "--rm",
                "-v",
                "{pex_project_dir}:/code".format(pex_project_dir=pex_project_dir()),
                "-v",
                "{test_script}:/test.sh".format(test_script=self.test_script),
                self.base_image,
                "/test.sh",
                "-c",
                python_code,
            ]
            + list(args)
        )



# Copyright 2021 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import json
import os
import re
import subprocess
from textwrap import dedent

import pytest

from pex.compatibility import PY2
from pex.dist_metadata import Constraint, Requirement
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.pip.version import PipVersion
from pex.resolve.locked_resolve import Artifact, LockedRequirement, LockedResolve, LockStyle
from pex.resolve.lockfile import json_codec
from pex.resolve.lockfile.json_codec import ParseError, PathMappingError
from pex.resolve.lockfile.model import Lockfile
from pex.resolve.path_mappings import PathMapping, PathMappings
from pex.resolve.resolved_requirement import ArtifactURL, Fingerprint, Pin
from pex.resolve.resolver_configuration import BuildConfiguration, ResolverVersion
from pex.sorted_tuple import SortedTuple
from pex.third_party.packaging import tags
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Container

    import attr  # vendor:skip
else:
    import pex.third_party.attr as attr


def test_roundtrip(tmpdir):
    # type: (Any) -> None

    lockfile = Lockfile.create(
        pex_version="1.2.3",
        style=LockStyle.STRICT,
        requires_python=(),
        target_systems=(),
        pip_version=PipVersion.VENDORED,
        resolver_version=ResolverVersion.PIP_2020,
        requirements=(
            Requirement.parse("ansicolors"),
            Requirement.parse("requests>=2; sys_platform == 'darwin'"),
        ),
        constraints=(Constraint.parse("ansicolors==1.1.8"),),
        allow_prereleases=True,
        build_configuration=BuildConfiguration(),
        transitive=False,
        excluded=(),
        overridden=(),
        locked_resolves=[
            LockedResolve(
                platform_tag=tags.Tag("cp36", "cp36m", "macosx_10_13_x86_64"),
                locked_requirements=SortedTuple(
                    [
                        LockedRequirement.create(
                            pin=Pin(
                                project_name=ProjectName("ansicolors"), version=Version("1.1.8")
                            ),
                            artifact=Artifact.from_url(
                                url="https://example.org/colors-1.1.8-cp36-cp36m-macosx_10_6_x86_64.whl",
                                fingerprint=Fingerprint(algorithm="blake256", hash="cafebabe"),
                            ),
                            additional_artifacts=(),
                        ),
                        LockedRequirement.create(
                            pin=Pin(project_name=ProjectName("requests"), version=Version("2.0.0")),
                            artifact=Artifact.from_url(
                                url="https://example.org/requests-2.0.0-py2.py3-none-any.whl",
                                fingerprint=Fingerprint(algorithm="sha256", hash="456"),
                            ),
                            additional_artifacts=(
                                Artifact.from_url(
                                    url="file://find-links/requests-2.0.0.tar.gz",
                                    fingerprint=Fingerprint(algorithm="sha512", hash="123"),
                                ),
                            ),
                        ),
                    ]
                ),
            ),
            LockedResolve(
                platform_tag=tags.Tag("cp37", "cp37m", "manylinux1_x86_64"),
                locked_requirements=SortedTuple(
                    [
                        LockedRequirement.create(
                            pin=Pin(
                                project_name=ProjectName("ansicolors"), version=Version("1.1.8")
                            ),
                            artifact=Artifact.from_url(
                                url="https://example.org/colors-1.1.8-cp37-cp37m-manylinux1_x86_64.whl",
                                fingerprint=Fingerprint(algorithm="md5", hash="hackme"),
                            ),
                            additional_artifacts=(),
                        ),
                    ]
                ),
            ),
        ],
    )
    assert lockfile == json_codec.loads(json.dumps(json_codec.as_json_data(lockfile)))

    with open(os.path.join(str(tmpdir), "lock.json"), "w") as fp:
        json.dump(json_codec.as_json_data(lockfile), fp)
    assert lockfile == json_codec.load(fp.name)


VALID_LOCK = """\
{
  "allow_builds": true,
  "allow_prereleases": false,
  "allow_wheels": true,
  "build_isolation": true,
  "constraints": [],
  "locked_resolves": [
    {
      "locked_requirements": [
        {
          "artifacts": [
            {
              "algorithm": "md5",
              "hash": "f357aa02db2466bc24ff1815cff1aeb3",
              "url": "file:///find-links/ansicolors-1.1.8-py2.py3-none-any.whl"
            },
            {
              "algorithm": "md5",
              "hash": "9ca7e2396ffa2e20af023c6b83ab7b14",
              "url": "http://localhost:9999/ansicolors-1.1.8.zip"
            }
          ],
          "project_name": "ansicolors",
          "requires_dists": [],
          "requires_python": null,
          "version": "1.1.8"
        }
      ],
      "platform_tag": [
        "cp39",
        "cp39",
        "manylinux_2_33_x86_64"
      ]
    }
  ],
  "pex_version": "2.1.50",
  "prefer_older_binary": false,
  "requirements": [
    "ansicolors"
  ],
  "requires_python": [],
  "resolver_version": "pip-legacy-resolver",
  "style": "sources",
  "transitive": true,
  "use_pep517": null
}
"""


@attr.s(frozen=True)
class PatchTool(object):
    tmpdir = attr.ib()  # type: str

    def apply(self, patch):
        # type: (str) -> str
        lock_file = os.path.join(self.tmpdir, "lock.json")
        with open(lock_file, "w") as fp:
            fp.write(VALID_LOCK)
        patch = dedent(
            """\
            --- a/{lock_file}	
            +++ b/{lock_file}
            {patch}
            """
        ).format(lock_file=os.path.basename(lock_file), patch=patch)
        process = subprocess.Popen(args=["git", "apply"], cwd=self.tmpdir, stdin=subprocess.PIPE)
        process.communicate(input=patch.encode("utf-8"))
        assert 0 == process.returncode, "Applying patch failed with exit code {}".format(
            process.returncode
        )
        with open(lock_file) as fp:
            return fp.read()


@pytest.fixture
def patch_tool(tmpdir):
    # type: (Any) -> PatchTool
    return PatchTool(str(tmpdir))


def parse_patched(
    patch_tool,  # type: PatchTool
    patch,  # type: str
    path_mappings=PathMappings(),  # type: PathMappings
):
    # type: (...) -> Lockfile
    return json_codec.loads(lockfile_contents=patch_tool.apply(patch), path_mappings=path_mappings)


def assert_parse_error(
    patch_tool,  # type: PatchTool
    patch,  # type: str
    match,  # type: str
    path_mappings=PathMappings(),  # type: PathMappings
):
    # type: (...) -> None
    with pytest.raises(ParseError, match=match):
        parse_patched(patch_tool, patch, path_mappings=path_mappings)


def test_load_invalid_json(patch_tool):
    # type: (PatchTool) -> None

    assert_parse_error(
        patch_tool,
        dedent(
            """\
            @@ -1,2 +1,2 @@
            -{
            +[
               "allow_builds": true,
            """
        ),
        "The lock file at <string> does not contain valid JSON:",
    )


def test_load_invalid_type(patch_tool):
    # type: (PatchTool) -> None

    assert_parse_error(
        patch_tool,
        dedent(
            """\
            @@ -1,3 +1,3 @@
             {
            -  "allow_builds": true,
            +  "allow_builds": 42,
               "allow_prereleases": false,
            """
        ),
        re.escape(
            "Expected '.[\"allow_builds\"]' in <string> to be of type bool but given int with "
            "value 42."
        ),
    )


def test_load_invalid_key_not_found(patch_tool):
    # type: (PatchTool) -> None

    assert_parse_error(
        patch_tool,
        dedent(
            """\
            @@ -11,3 +11,2 @@
                         {
            -              "algorithm": "md5",
                           "hash": "f357aa02db2466bc24ff1815cff1aeb3",
            """
        ),
        match=re.escape(
            "The object at '.locked_resolves[0][0][\"artifacts\"][0]' in <string> did not have the "
            "expected key 'algorithm'."
        ),
    )
    assert_parse_error(
        patch_tool,
        dedent(
            """\
            @@ -12,3 +12,3 @@
                           "algorithm": "md5",
            -              "hash": "f357aa02db2466bc24ff1815cff1aeb3",
            +              "HASH": "f357aa02db2466bc24ff1815cff1aeb3",
                           "url": "file:///find-links/ansicolors-1.1.8-py2.py3-none-any.whl"
            """
        ),
        match=re.escape(
            "The object at '.locked_resolves[0][0][\"artifacts\"][0]' in <string> did not have the "
            "expected key 'hash'."
        ),
    )


FORMAT_ARGS = dict(
    str_type="unicode" if PY2 else "str",
    str_prefix="u" if PY2 else "",
)


def test_load_invalid_parent_not_json_object(patch_tool):
    # type: (PatchTool) -> None

    assert_parse_error(
        patch_tool,
        dedent(
            """\
            @@ -10,7 +10,3 @@
                       "artifacts": [
            -            {
            -              "algorithm": "md5",
            -              "hash": "f357aa02db2466bc24ff1815cff1aeb3",
            -              "url": "file:///find-links/ansicolors-1.1.8-py2.py3-none-any.whl"
            -            },
            +            "foo",
                         {
            """
        ),
        match=re.escape(
            'Cannot retrieve \'.locked_resolves[0][0]["artifacts"][0]["url"]\' in <string> '
            "because '.locked_resolves[0][0][\"artifacts\"][0]' is not a JSON object but a "
            "{str_type} with value foo.".format(**FORMAT_ARGS)
        ),
    )


def test_load_invalid_requirement(patch_tool):
    # type: (PatchTool) -> None

    assert_parse_error(
        patch_tool,
        dedent(
            """\
            @@ -38,3 +38,4 @@
               "requirements": [
            -    "ansicolors"
            +    "ansicolors",
            +    "@invalid requirement"
               ],
            """
        ),
        match=re.escape("The requirement string at '.requirements[1]' is invalid:"),
    )

    assert_parse_error(
        patch_tool,
        dedent(
            """\
            @@ -5,3 +5,3 @@
               "build_isolation": true,
            -  "constraints": [],
            +  "constraints": ["@invalid requirement"],
               "locked_resolves": [
            """
        ),
        match=re.escape("The requirement string at '.constraints[0]' is invalid:"),
    )

    assert_parse_error(
        patch_tool,
        dedent(
            """\
            @@ -23,3 +23,6 @@
                       "project_name": "ansicolors",
            -          "requires_dists": [],
            +          "requires_dists": [
            +            "valid_requirement",
            +            "@invalid_requirement"
            +          ],
                       "requires_python": null,
            """
        ),
        match=re.escape(
            "The requirement string at '.locked_resolves[0][0][\"requires_dists\"][1]' is invalid:"
        ),
    )


def test_load_invalid_requires_python(patch_tool):
    # type: (PatchTool) -> None

    assert_parse_error(
        patch_tool,
        dedent(
            """\
            --- lock.orig.json	2022-01-23 16:25:23.099399463 -0800
            +++ lock.json	2022-01-23 16:39:35.547488554 -0800
            @@ -24,3 +24,3 @@
                       "requires_dists": [],
            -          "requires_python": null,
            +          "requires_python": "@invalid specifier",
                       "version": "1.1.8"
            """
        ),
        match=re.escape(
            "The version specifier at '.locked_resolves[0][0][\"requires_python\"]' is invalid:"
        ),
    )


def test_load_invalid_resolver_version(patch_tool):
    # type: (PatchTool) -> None

    assert_parse_error(
        patch_tool,
        dedent(
            """\
            @@ -39,3 +39,3 @@
               "requires_python": [],
            -  "resolver_version": "pip-legacy-resolver",
            +  "resolver_version": "apache-ivy",
               "style": "sources",
            """
        ),
        match=re.escape("The '.[\"resolver_version\"]' is invalid: "),
    )


def test_load_invalid_style(patch_tool):
    # type: (PatchTool) -> None

    assert_parse_error(
        patch_tool,
        dedent(
            """\
            @@ -41,3 +41,3 @@
               "resolver_version": "pip-legacy-resolver",
            -  "style": "sources",
            +  "style": "foo",
               "transitive": true,
            """
        ),
        match=re.escape("The '.[\"style\"]' is invalid: "),
    )


def test_load_invalid_platform_tag(patch_tool):
    # type: (PatchTool) -> None

    assert_parse_error(
        patch_tool,
        dedent(
            """\
            @@ -28,3 +28,2 @@
                   "platform_tag": [
            -        "cp39",
                     "cp39",
            """
        ),
        match=re.escape(
            "The tag at '.locked_resolves[0][\"platform_tag\"]' must have 3 string components. "
            "Given 2 with types [{str_type}, {str_type}]: [{str_prefix}'cp39', "
            "{str_prefix}'manylinux_2_33_x86_64']".format(**FORMAT_ARGS)
        ),
    )

    assert_parse_error(
        patch_tool,
        dedent(
            """\
            @@ -29,3 +29,3 @@
                     "cp39",
            -        "cp39",
            +        42,
                     "manylinux_2_33_x86_64"
            """
        ),
        match=re.escape(
            "The tag at '.locked_resolves[0][\"platform_tag\"]' must have 3 string components. "
            "Given 3 with types [{str_type}, int, {str_type}]: [{str_prefix}'cp39', 42, "
            "{str_prefix}'manylinux_2_33_x86_64']".format(**FORMAT_ARGS)
        ),
    )


def test_load_invalid_no_artifacts(patch_tool):
    # type: (PatchTool) -> None

    assert_parse_error(
        patch_tool,
        dedent(
            """\
            @@ -10,12 +10,2 @@
                       "artifacts": [
            -            {
            -              "algorithm": "md5",
            -              "hash": "f357aa02db2466bc24ff1815cff1aeb3",
            -              "url": "file:///find-links/ansicolors-1.1.8-py2.py3-none-any.whl"
            -            },
            -            {
            -              "algorithm": "md5",
            -              "hash": "9ca7e2396ffa2e20af023c6b83ab7b14",
            -              "url": "http://localhost:9999/ansicolors-1.1.8.zip"
            -            }
                       ],
            """
        ),
        match=re.escape(
            "Expected '.locked_resolves[0][0]' in <string> to have at least one artifact."
        ),
    )


def test_path_mappings_required_on_parse(patch_tool):
    # type: (PatchTool) -> None

    patch = dedent(
        """\
        @@ -35,2 +35,6 @@
           ],
        +  "path_mappings": {
        +    "FOO": "The fooey find links repo path on the local machine.",
        +    "BAR": null
        +  },
           "pex_version": "2.1.50",
        """
    )

    def assert_mapping_error(
        unspecified_paths,  # type: Container[str]
        *path_mappings  # type: PathMapping
    ):
        with pytest.raises(PathMappingError) as exec_info:
            parse_patched(patch_tool, patch, path_mappings=PathMappings(path_mappings))
        assert (
            PathMappingError(
                required_path_mappings={
                    "BAR": None,
                    "FOO": "The fooey find links repo path on the local machine.",
                },
                unspecified_paths=unspecified_paths,
            )
            == exec_info.value
        )

    assert_mapping_error({"BAR", "FOO"})
    assert_mapping_error({"FOO"}, PathMapping(name="BAR", path="/tmp/bar"))
    assert_mapping_error({"BAR"}, PathMapping(name="FOO", path="/tmp/foo"))

    assert (
        parse_patched(
            patch_tool,
            patch,
            path_mappings=PathMappings(
                (PathMapping(name="FOO", path="/tmp/foo"), PathMapping(name="BAR", path="/tmp/bar"))
            ),
        )
        is not None
    )


def test_path_mappings_round_trip():
    # type: () -> None

    lock_file = json_codec.loads(VALID_LOCK)

    data = json_codec.as_json_data(
        lock_file,
        PathMappings(
            (
                PathMapping(
                    path="/find-links",
                    name="FL",
                    description="Our NFS find-links repo local mount path.",
                ),
            )
        ),
    )
    locked_resolves = data["locked_resolves"]
    assert 1 == len(locked_resolves)
    locked_requirements = locked_resolves[0]["locked_requirements"]
    assert 1 == len(locked_requirements)
    artifacts = locked_requirements[0]["artifacts"]
    assert 2 == len(artifacts)
    assert "file://${FL}/ansicolors-1.1.8-py2.py3-none-any.whl" == artifacts[0]["url"]
    assert "http://localhost:9999/ansicolors-1.1.8.zip" == artifacts[1]["url"]

    canonicalized_lock = json.dumps(data)
    with pytest.raises(PathMappingError) as exc_info:
        json_codec.loads(canonicalized_lock)
    assert (
        PathMappingError(
            required_path_mappings={"FL": "Our NFS find-links repo local mount path."},
            unspecified_paths={"FL"},
        )
        == exc_info.value
    )

    assert lock_file == json_codec.loads(
        canonicalized_lock,
        path_mappings=PathMappings((PathMapping(name="FL", path="/find-links"),)),
    ), (
        "Expected a round-trip through canonicalize / reify with the same mappings to return the "
        "same lock contents."
    )

    lock_file2 = json_codec.loads(
        canonicalized_lock,
        path_mappings=PathMappings((PathMapping(name="FL", path="/at/another/path/find-links"),)),
    )
    assert (
        lock_file2 != lock_file
    ), "Expected path roots to be reified to different values than we started with."
    assert (
        ArtifactURL.parse(
            "file:///at/another/path/find-links/ansicolors-1.1.8-py2.py3-none-any.whl"
        )
        == lock_file2.locked_resolves[0].locked_requirements[0].artifact.url
    )



# Copyright 2022 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import re
from collections import defaultdict

import pytest

from pex.common import safe_mkdtemp
from pex.dist_metadata import DistributionType, Requirement
from pex.interpreter import PythonInterpreter
from pex.pep_427 import InstallableType
from pex.pex_builder import PEXBuilder
from pex.pex_info import PexInfo
from pex.platforms import Platform
from pex.resolve.configured_resolver import ConfiguredResolver
from pex.resolve.pex_repository_resolver import resolve_from_pex
from pex.resolve.resolver_configuration import PipConfiguration
from pex.resolve.resolvers import ResolveResult, Unsatisfiable
from pex.resolver import resolve
from pex.targets import Targets
from pex.typing import TYPE_CHECKING, cast
from testing import IS_LINUX, PY27, PY310, ensure_python_interpreter

if TYPE_CHECKING:
    from typing import DefaultDict, Iterable, Optional, Set


def create_pex_repository(
    interpreters=None,  # type: Optional[Iterable[PythonInterpreter]]
    platforms=None,  # type: Optional[Iterable[Platform]]
    requirements=None,  # type: Optional[Iterable[str]]
    requirement_files=None,  # type: Optional[Iterable[str]]
    constraint_files=None,  # type: Optional[Iterable[str]]
    manylinux=None,  # type: Optional[str]
    result_type=InstallableType.INSTALLED_WHEEL_CHROOT,  # type: InstallableType.Value
):
    # type: (...) -> str
    pex_builder = PEXBuilder()
    pex_builder.info.deps_are_wheel_files = result_type is InstallableType.WHEEL_FILE
    for resolved_dist in resolve(
        targets=Targets(
            interpreters=tuple(interpreters) if interpreters else (),
            platforms=tuple(platforms) if platforms else (),
            assume_manylinux=manylinux,
        ),
        requirements=requirements,
        requirement_files=requirement_files,
        constraint_files=constraint_files,
        resolver=ConfiguredResolver(PipConfiguration()),
        result_type=result_type,
    ).distributions:
        pex_builder.add_distribution(resolved_dist.distribution)
        for direct_req in resolved_dist.direct_requirements:
            pex_builder.add_requirement(direct_req)
    pex_builder.freeze()
    return os.path.realpath(cast(str, pex_builder.path()))


def create_constraints_file(*requirements):
    # type: (*str) -> str
    constraints_file = os.path.join(safe_mkdtemp(), "constraints.txt")
    with open(constraints_file, "w") as fp:
        for requirement in requirements:
            fp.write(requirement + os.linesep)
    return constraints_file


@pytest.fixture(scope="module")
def py27():
    # type: () -> PythonInterpreter
    return PythonInterpreter.from_binary(ensure_python_interpreter(PY27))


@pytest.fixture(scope="module")
def py310():
    # type: () -> PythonInterpreter
    return PythonInterpreter.from_binary(ensure_python_interpreter(PY310))


@pytest.fixture(scope="module")
def macosx():
    # type: () -> Platform
    return Platform.create("macosx-10.13-x86_64-cp-36-m")


@pytest.fixture(scope="module")
def linux():
    # type: () -> Platform
    return Platform.create("linux-x86_64-cp-36-m")


@pytest.fixture(scope="module")
def manylinux():
    # type: () -> Optional[str]
    return None if IS_LINUX else "manylinux2014"


@pytest.fixture(scope="module")
def foreign_platform(
    macosx,  # type: Platform
    linux,  # type: Platform
):
    # type: (...) -> Platform
    return macosx if IS_LINUX else linux


@pytest.fixture(
    scope="module",
    params=[
        pytest.param(installable_type, id=installable_type.value)
        for installable_type in InstallableType.values()
    ],
)
def pex_repository(
    py27,  # type: PythonInterpreter
    py310,  # type: PythonInterpreter
    foreign_platform,  # type: Platform
    manylinux,  # type: Optional[str]
    request,  # type: pytest.FixtureRequest
):
    # type (...) -> str

    constraints_file = create_constraints_file(
        # The 2.25.1 release of requests constrains urllib3 to <1.27,>=1.21.1 and picks 1.26.2 on
        # its own as of this writing.
        "urllib3==1.26.1",
        # The 22.0.0 release of pyOpenSSL drops support for Python 2.7; so we pin lower.
        "pyOpenSSL<22",
        # The 2022 and later releases only support Python>=3.6; so we pin lower.
        "certifi<2022",
        # The 2.22 release of pycparser drops support for Python 2.7.
        "pycparser<2.22",
    )

    return create_pex_repository(
        interpreters=[py27, py310],
        platforms=[foreign_platform],
        requirements=["requests[security,socks]==2.25.1"],
        constraint_files=[constraints_file],
        manylinux=manylinux,
        result_type=request.param,
    )


def test_resolve_from_pex(
    pex_repository,  # type: str
    py27,  # type: PythonInterpreter
    py310,  # type: PythonInterpreter
    foreign_platform,  # type: Platform
    manylinux,  # type: Optional[str]
):
    # type: (...) -> None
    pex_info = PexInfo.from_pex(pex_repository)
    direct_requirements = pex_info.requirements
    assert 1 == len(direct_requirements)

    def assert_resolve_result(
        result,  # type: ResolveResult
        expected_result_type,  # type: InstallableType.Value
    ):
        # type: (...) -> None

        assert expected_result_type is result.type
        expected_dist_type = (
            DistributionType.WHEEL
            if expected_result_type is InstallableType.WHEEL_FILE
            else DistributionType.INSTALLED
        )

        distribution_locations_by_key = defaultdict(set)  # type: DefaultDict[str, Set[str]]
        for resolved_distribution in result.distributions:
            assert expected_dist_type is resolved_distribution.distribution.type
            distribution_locations_by_key[resolved_distribution.distribution.key].add(
                resolved_distribution.distribution.location
            )

        assert {
            os.path.basename(location)
            for locations in distribution_locations_by_key.values()
            for location in locations
        } == set(pex_info.distributions.keys()), (
            "Expected to resolve the same full set of distributions from the pex repository as make "
            "it up when using the same requirements."
        )

        assert "requests" in distribution_locations_by_key
        assert 1 == len(distribution_locations_by_key["requests"])

        assert "pysocks" in distribution_locations_by_key
        assert 2 == len(distribution_locations_by_key["pysocks"]), (
            "PySocks has a non-platform-specific Python 2.7 distribution and a non-platform-specific "
            "Python 3 distribution; so we expect to resolve two distributions - one covering "
            "Python 2.7 and one covering local Python 3.6 and our cp36 foreign platform."
        )

        assert "cryptography" in distribution_locations_by_key
        assert 3 == len(distribution_locations_by_key["cryptography"]), (
            "The cryptography requirement of the security extra is platform specific; so we expect a "
            "unique distribution to be resolved for each of the three distribution targets"
        )

    assert_resolve_result(
        resolve_from_pex(
            pex=pex_repository,
            requirements=direct_requirements,
            targets=Targets(
                interpreters=(py27, py310),
                platforms=(foreign_platform,),
                assume_manylinux=manylinux,
            ),
        ),
        expected_result_type=InstallableType.INSTALLED_WHEEL_CHROOT,
    )

    if pex_info.deps_are_wheel_files:
        assert_resolve_result(
            resolve_from_pex(
                pex=pex_repository,
                requirements=direct_requirements,
                targets=Targets(
                    interpreters=(py27, py310),
                    platforms=(foreign_platform,),
                    assume_manylinux=manylinux,
                ),
                result_type=InstallableType.WHEEL_FILE,
            ),
            expected_result_type=InstallableType.WHEEL_FILE,
        )
    else:
        with pytest.raises(
            Unsatisfiable,
            match=(
                r"Cannot resolve \.whl files from PEX at {pex}; its dependencies are in the form "
                r"of pre-installed wheel chroots\.".format(pex=re.escape(pex_repository))
            ),
        ):
            resolve_from_pex(
                pex=pex_repository,
                requirements=direct_requirements,
                targets=Targets(
                    interpreters=(py27, py310),
                    platforms=(foreign_platform,),
                    assume_manylinux=manylinux,
                ),
                result_type=InstallableType.WHEEL_FILE,
            )


def test_resolve_from_pex_subset(
    pex_repository,  # type: str
    foreign_platform,  # type: Platform
    manylinux,  # type: Optional[str]
):
    # type: (...) -> None

    result = resolve_from_pex(
        pex=pex_repository,
        requirements=["cffi"],
        targets=Targets(
            platforms=(foreign_platform,),
            assume_manylinux=manylinux,
        ),
    )

    assert {"cffi", "pycparser"} == {
        resolved_distribution.distribution.project_name
        for resolved_distribution in result.distributions
    }


def test_resolve_from_pex_not_found(
    pex_repository,  # type: str
    py310,  # type: PythonInterpreter
):
    # type: (...) -> None

    with pytest.raises(Unsatisfiable) as exec_info:
        resolve_from_pex(
            pex=pex_repository,
            requirements=["pex"],
            targets=Targets(
                interpreters=(py310,),
            ),
        )
    assert "A distribution for pex could not be resolved for {py310_exe}.".format(
        py310_exe=py310.binary
    ) in str(exec_info.value)

    with pytest.raises(Unsatisfiable) as exec_info:
        resolve_from_pex(
            pex=pex_repository,
            requirements=["requests==1.0.0"],
            targets=Targets(
                interpreters=(py310,),
            ),
        )
    message = str(exec_info.value)
    assert (
        "Failed to resolve requirements from PEX environment @ {}".format(pex_repository) in message
    )
    assert "Needed {} compatible dependencies for:".format(py310.platform.tag) in message
    assert "1: requests==1.0.0" in message
    assert "But this pex only contains:" in message
    assert "requests-2.25.1-py2.py3-none-any.whl" in message


def test_resolve_from_pex_intransitive(
    pex_repository,  # type: str
    py27,  # type: PythonInterpreter
    py310,  # type: PythonInterpreter
    foreign_platform,  # type: Platform
    manylinux,  # type: Optional[str]
):
    # type: (...) -> None

    resolved_distributions = resolve_from_pex(
        pex=pex_repository,
        requirements=["requests"],
        transitive=False,
        targets=Targets(
            interpreters=(py27, py310),
            platforms=(foreign_platform,),
            assume_manylinux=manylinux,
        ),
    ).distributions
    assert 3 == len(
        resolved_distributions
    ), "Expected one resolved distribution per distribution target."
    assert 1 == len(
        frozenset(
            resolved_distribution.distribution.location
            for resolved_distribution in resolved_distributions
        )
    ), (
        "Expected one underlying resolved universal distribution usable on Linux and macOs by "
        "both Python 2.7 and Python 3.6."
    )
    for resolved_distribution in resolved_distributions:
        assert (
            Requirement.parse("requests==2.25.1")
            == resolved_distribution.distribution.as_requirement()
        )
        assert 1 == len(resolved_distribution.direct_requirements)
        assert Requirement.parse("requests") == resolved_distribution.direct_requirements[0]


def test_resolve_from_pex_constraints(
    pex_repository,  # type: str
    py27,  # type: PythonInterpreter
):
    # type: (...) -> None

    with pytest.raises(Unsatisfiable) as exec_info:
        resolve_from_pex(
            pex=pex_repository,
            requirements=["requests"],
            constraint_files=[create_constraints_file("urllib3==1.26.2")],
            targets=Targets(
                interpreters=(py27,),
            ),
        )
    message = str(exec_info.value)
    assert "The following constraints were not satisfied by " in message
    assert " resolved from {}:".format(pex_repository) in message
    assert "urllib3==1.26.2" in message


def test_resolve_from_pex_ignore_errors(
    pex_repository,  # type: str
    py27,  # type: PythonInterpreter
):
    # type: (...) -> None

    # See test_resolve_from_pex_constraints above for the failure this would otherwise cause.
    result = resolve_from_pex(
        pex=pex_repository,
        requirements=["requests"],
        constraint_files=[create_constraints_file("urllib3==1.26.2")],
        targets=Targets(
            interpreters=(py27,),
        ),
        ignore_errors=True,
    )
    resolved_distributions_by_key = {
        resolved_distribution.distribution.project_name: resolved_distribution.distribution.as_requirement()
        for resolved_distribution in result.distributions
    }
    assert len(resolved_distributions_by_key) > 1, "We should resolve at least requests and urllib3"
    assert "requests" in resolved_distributions_by_key
    assert Requirement.parse("urllib3==1.26.1") == resolved_distributions_by_key["urllib3"]



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os

from pex.common import CopyMode
from pex.typing import TYPE_CHECKING
from pex.venv.virtualenv import Virtualenv

if TYPE_CHECKING:
    from typing import Iterable, Optional, Set


def assert_venv_site_packages_copy_mode(
    venv_dir,  # type: str
    expected_copy_mode,  # type: CopyMode.Value
    expected_files=None,  # type: Optional[Iterable[str]]
):
    # type: (...) -> None

    site_packages_files = set()  # type: Set[str]
    site_packages_dir = os.path.realpath(Virtualenv(venv_dir).site_packages_dir)
    for root, dirs, files in os.walk(site_packages_dir):
        # Metadata files are always copied for inscrutable historical reasons; so we skip
        # checking those.
        dirs[:] = [d for d in dirs if not d.endswith(".dist-info")]
        for f in files:
            if f == "PEX_EXTRA_SYS_PATH.pth":
                continue
            file_path = os.path.join(root, f)
            if expected_copy_mode is CopyMode.SYMLINK:
                assert os.path.islink(file_path)
            else:
                assert not os.path.islink(file_path)
                nlink = os.stat(file_path).st_nlink
                if expected_copy_mode is CopyMode.COPY:
                    assert 1 == nlink, "Expected {file} to have 1 link but found {nlink}".format(
                        file=file_path, nlink=nlink
                    )
                else:
                    assert (
                        nlink > 1
                    ), "Expected {file} to have more than 1 link but found {nlink}".format(
                        file=file_path, nlink=nlink
                    )
            site_packages_files.add(file_path)

    if expected_files:
        expected_files = set(os.path.join(site_packages_dir, f) for f in expected_files)
        assert expected_files == site_packages_files, (
            "Expected venv site-packages dir to contain these files:\n"
            "{expected_files}\n"
            "\n"
            "Found these files:\n"
            "{actual_files}"
        ).format(
            expected_files="\n".join(sorted(expected_files)),
            actual_files="\n".join(sorted(site_packages_files)),
        )



# Copyright 2021 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from pex.typing import TYPE_CHECKING, Generic, overload

if TYPE_CHECKING:
    from typing import Any, Iterable, Iterator, Optional, Protocol, TypeVar, Union

    from typing_extensions import SupportsIndex

    _T = TypeVar("_T")

    class _Comparable(Protocol):
        def __lt__(self, other):
            # type: (Any) -> bool
            pass

    class _TComparator(Protocol):
        def __call__(self, item):
            # type: (_T) -> _Comparable
            pass


class SortedTuple(Generic["_T"], tuple):
    @overload
    def __new__(cls):
        # type: () -> SortedTuple[Any]
        pass

    @overload
    def __new__(
        cls,
        iterable,  # type: Iterable[_T]
        key=None,  # type: None
        reverse=False,  # type: bool
    ):
        # type: (...) -> SortedTuple[_T]
        pass

    @overload
    def __new__(
        cls,
        iterable,  # type: Iterable[_T]
        key,  # type: _TComparator
        reverse=False,  # type: bool
    ):
        # type: (...) -> SortedTuple[_T]
        pass

    def __new__(
        cls,
        iterable=(),  # type: Iterable[_T]
        key=None,  # type: Optional[_TComparator]
        reverse=False,  # type: bool
    ):
        # type: (...) -> SortedTuple[_T]
        return super(SortedTuple, cls).__new__(
            cls,
            # There appears to be no way to express that _T should be comparable if no key function
            # is passed, but otherwise should be comparable.
            sorted(iterable, key=key, reverse=reverse),  # type: ignore[arg-type, type-var]
        )

    @overload
    def __getitem__(self, index):
        # type: (SupportsIndex) -> _T
        pass

    @overload
    def __getitem__(self, slice_spec):
        # type: (slice) -> SortedTuple[_T]
        pass

    def __getitem__(self, item):
        # type: (Union[SupportsIndex, slice]) -> Union[_T, SortedTuple[_T]]

        # MyPy `--python-version 2.7` does not understand SupportsIndex and the bifurcated return
        # type does not appear to be expressible.
        return tuple.__getitem__(self, item)  # type: ignore[index, return-value]

    def __iter__(self):
        # type: () -> Iterator[_T]
        return tuple.__iter__(self)
